# Copyright (c) 2025, InfinityQ Technology, Inc.

import abc
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Mapping, Optional, Tuple
from urllib.parse import urljoin

from pydantic import BaseModel, ValidationError
import requests

from titanq import errors
from titanq._api.model.problem import Problem


_AUTHORIZATION_HEADER = "Authorization"
_TRACE_ID_HEADER = 'X-Correlation-Id'
_USER_AGENT_HEADER = 'User-Agent'


@dataclass
class HttpRequest:
    method: str
    url: str
    headers: Dict[str, str] = field(default_factory=dict)
    body: Optional[BaseModel] = None
    query_param: Optional[BaseModel] = None


@dataclass
class HttpMetadata:
    status_code: int
    headers: Mapping[str, str]


class HttpClient(abc.ABC):
    @abc.abstractmethod
    def do(self, http_request: HttpRequest) -> Tuple[bytes, HttpMetadata]: ...


class HTTPMethod(str, Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"


class BasicHttpClient(HttpClient):

    def __init__(self, base_server_address: str) -> None:
        self._base_server_address = base_server_address

    def do(self, http_request: HttpRequest) -> Tuple[bytes, HttpMetadata]:
        response = requests.request(
            method=http_request.method,
            headers=http_request.headers,
            url=urljoin(self._base_server_address, http_request.url),
            json=http_request.body.model_dump(exclude_unset=True) if http_request.body else None,
            params=http_request.query_param.model_dump(exclude_unset=True) if http_request.query_param else None,
        )

        # check if the result is a problem
        try:
            Problem.model_validate_json(response.content).raise_()
        except ValidationError:
            pass

        # if not a problem
        try:
            return response.content, HttpMetadata(
                status_code=response.status_code,
                headers=response.headers
            )
        except ValidationError:
            raise errors.UnexpectedServerResponseError(
                 f"Failed to parse the response from the TitanQ server. The raw response is: {response.content}"
            )


class HttpClientWithDefaultHeader(HttpClient):
    """HttpClient decorator that add an header on each request"""
    def __init__(self, client: HttpClient, key: str, value: str) -> None:
        self._key = key
        self._value = value
        self._client = client


    def do(self, http_request: HttpRequest) -> Tuple[bytes, HttpMetadata]:
        http_request.headers[self._key] = self._value
        return self._client.do(http_request)


class HttpClientWithAuthorization(HttpClientWithDefaultHeader):
    def __init__(self, client: HttpClient, api_key: str) -> None:
        super().__init__(client, _AUTHORIZATION_HEADER, api_key)


class HttpClientWithSession(HttpClientWithDefaultHeader):
    def __init__(self, client: HttpClient, trace_id: str) -> None:
        super().__init__(client, _TRACE_ID_HEADER, trace_id)


class HttpClientWithUserAgent(HttpClientWithDefaultHeader):
    def __init__(self, client: HttpClient, product_name: str, product_version: str) -> None:
        super().__init__(
            client,
            _USER_AGENT_HEADER,
            HttpClientWithUserAgent._user_agent_str(product_name, product_version)
        )

    @classmethod
    def _user_agent_str(_cls, product_name: str, product_version: str):
        request_user_agent = requests.utils.default_headers().get(_USER_AGENT_HEADER, '')
        user_agent = f"{product_name}/{product_version} " + request_user_agent
        return user_agent.rstrip()


# Copyright (c) 2024, InfinityQ Technology, Inc.

# core functions
from .file_parser import parse_from_mps, export_to_mps
from .options import MPSParseOptions

# for typing purpose
from .model import (
    MPSBounds,
    MPSColumns,
    MPSName,
    MPSObject,
    MPSQuadobj,
    MPSRanges,
    MPSRhs,
    MPSRow
)
from .utils import BoundsType, ColumnsType, RowsType


# Copyright (c) 2025, InfinityQ Technology, Inc.

from abc import ABC, abstractmethod
from io import BytesIO
from typing import Optional

from titanq._api.model.solve import Manifest, Request


class StorageClient(ABC):

    @abstractmethod
    def init(self) -> None:
        """Initialize the storage client"""

    @abstractmethod
    def request_input_builder(
        self,
        bias: BytesIO,
        variable_bounds: BytesIO,
        weights: Optional[BytesIO] = None,
        constraint_weights: Optional[BytesIO] = None,
        constraint_bounds: Optional[BytesIO] = None,
        quad_constraint_weights: Optional[BytesIO] = None,
        quad_constraint_bounds: Optional[BytesIO] = None,
        quad_constraint_linear_weights: Optional[BytesIO] = None,
        manifest: Optional[Manifest] = None,
    ) -> Request.Input:
        """Builds the input request"""
        pass

    @abstractmethod
    def request_output_builder(self) -> Request.Output:
        """Builds the output request"""
        pass


# Copyright (c) 2024, InfinityQ Technology, Inc.
"""
This module provides utilities for symbolic mathematics and optimization, leveraging the SymEngine library. It streamlines the representation,
manipulation, and transformation of symbolic expressions and equations into formats suitable for titanQ.

Responsibilities:
------------------
- Encapsulating SymEngine objects in a user-friendly interface to enhance usability.
- Enabling representation and manipulation of symbolic expressions via the `Expression` class.
- Facilitating the creation and conversion of symbolic equations (`Equation` class) into constraints for optimization.
- Expanding symbolic terms into their constituent variables and coefficients.
- Generating matrices for TitanQ (e.g., bias, weight, constraint mask) from expressions and equations.

Why Use This Module:
---------------------
- Provides a fast and intuitive way to define and manipulate expressions and equations.
- Incorporates warnings and safeguards to assist users in debugging their symbolic formulations.

Potential Enhancements:
------------------------
- **Support for Additional Equation Types**: Extend capabilities to include quadratic or higher-degree constraints.
- **Enhanced Warnings and Safeguards**: Introduce more detailed messages for unsupported features and edge cases.

Limitations:
------------
1. **Avoid Generic Numerical Solvers**: This module is dedicated to symbolic manipulation and preparing optimization tasks, not for solving numerical equations directly.
2. **Exclude Non-Symbolic Utilities**: Maintain a clear focus on symbolic mathematics and optimization-related functionalities.

Example Usage:
---------------
```python
# Create a model
from titanq import Model
model = Model()

# Create variable vectors
x = model.add_variable_vector('x', 10)
y = model.add_variable_vector('y', 10)

# Define an objective function using an expression
model.set_objective_expression(sum(x + y) + y**2)

# Add a constraint from an expression
model.add_constraint_from_expression(x[2] > y[4])
```
"""


from collections import defaultdict
from typing import Any, Dict, NamedTuple, Optional, Tuple, Union
import warnings

import numpy as np
from numpy.typing import NDArray
import symengine as se
import symengine.lib.symengine_wrapper as se_type

from titanq._model.array.arraybuilder.index_value_array_builder import IndexValueArrayBuilder

from ..errors import ContradictoryExpressionError, TautologicalExpressionError
from .variables import Variable


class ExpressionComponent(NamedTuple):
    constant: float
    linear: NDArray
    quadratic: Optional[IndexValueArrayBuilder]


class ExpandedVariableList:
    """
    A class representing an expanded list of variables that can be safely used as dictionary keys.

    The `ExpandedVariableList` ensures that instances are hashable and can be compared reliably.
    This makes it suitable for use in scenarios where a list-like structure must serve as a key in a dictionary.

    Features:
    - Ensures immutability to guarantee consistent hash values.
    - Implements custom equality and hashing to allow comparison based on the content of the list.
    """
    def __init__(self, *variables: Variable) -> None:
        # Save _variables as tuple so it can be easily hashed and compared later.
        self._variables = tuple(sorted(variables, key=lambda v: v.problem_index()))


    def __hash__(self):
        return hash(self._variables)


    def __eq__(self, other):
        if isinstance(other, ExpandedVariableList):
            return self._variables == other._variables
        return False


    def __repr__(self):
        return repr(self._variables)


    def __iter__(self):
        return iter(self._variables)


    def __len__(self):
        return len(self._variables)


    def __getitem__(self, index):
        return self._variables[index]



class SymEngineObjectWrapper():
    """
    A wrapper class for encapsulating SymEngine objects, providing a convenient interface for representation
    and ensuring type safety.

    This class serves as a common parent for all mathematical objects, primarily to support the _get_se_object function.
    """

    def __init__(self, se_object: se_type.Basic):
        if not isinstance(se_object, se_type.Basic):
            raise TypeError(f"Expected an object of type symengine.Basic, got {type(se_object)}.")

        self._se_object = se_object


    def __repr__(self) -> str:
        return repr(self._se_object)


    def __str__(self) -> str:
        return str(self._se_object)

class Expression(SymEngineObjectWrapper):
    """
    A class for representing and manipulating mathematical expressions based on SymEngine.
    """

    def expand(self) -> 'Expression':
        """
        Expand the current expression algebraically.

        Returns
        -------
            Expression: A new expanded expression.
        """
        return Expression(se.expand(self._se_object))


    def terms(self) -> Dict[ExpandedVariableList, float]:
        """
        Decompose the expression into individual terms and their coefficients.

        Returns
        -------
            A mapping of variable tuples to their coefficients.
        """
        d = defaultdict(lambda: 0)
        for term, coeff in self._se_object.as_coefficients_dict().items():
            vars = _expand_term(term)
            d[vars] += coeff
        return d


    def split_into_component(self, n_variable: int) -> ExpressionComponent:
        """
        Convert the expression into constant, bias, and weight matrices.

        Returns
        -------
            - Constant term,
            - Bias vector,
            - Weight matrix (or None if no quadratic terms are present).

        Raise
        ------
            - ValueError: If n_variable is n <= 0.
            - ValueError: If The equation degree exceeds 2.
        """
        if n_variable <= 0:
            raise ValueError("Cannot generate matrices from expression for a non-positive number of variables")

        expr = self.expand()

        const: float = 0
        bias = np.zeros(n_variable, dtype=np.float32)
        weights = IndexValueArrayBuilder((n_variable, n_variable))

        for vars, coeffs in expr.terms().items():
            n_vars = len(vars)
            coeffs = float(coeffs)  # Coefficients must be numerical, not SymEngine type.

            if n_vars == 0:
                const += coeffs
            elif n_vars == 1:
                bias[vars[0].problem_index()] += coeffs
            elif n_vars == 2:
                index1, index2 = vars[0].problem_index(), vars[1].problem_index()
                weights.append(index1, index2, coeffs)
                weights.append(index2, index1, coeffs)
            else:
                raise ValueError("Currently only equations of degree 2 or lower are supported.")

        if weights.is_empty():
            return const, bias, None

        return const, bias, weights


    def __neg__(self) -> 'Expression':
        return Expression(-self._se_object)


    def __add__(self, other) -> 'Expression':
        return Expression(self._se_object + _get_se_object(other))


    def __sub__(self, other) -> 'Expression':
        return Expression(self._se_object - _get_se_object(other))


    def __mul__(self, other) -> 'Expression':
        return Expression(self._se_object * _get_se_object(other))


    def __pow__(self, other) -> 'Expression':
        return Expression(self._se_object ** other)


    def __radd__(self, other) -> 'Expression':
        return Expression(_get_se_object(other) + self._se_object)


    def __rsub__(self, other) -> 'Expression':
        return Expression(_get_se_object(other) - self._se_object)


    def __rmul__(self, other) -> 'Expression':
        return Expression(_get_se_object(other) * self._se_object)


    def __eq__(self, other) -> 'Equation':
        return Equation(se.Equality(self._se_object, _get_se_object(other)))


    def __ne__(self, other) -> 'Equation':
        return Equation(se.Unequality(self._se_object, _get_se_object(other)))


    def __le__(self, other) -> 'Equation':
        return Equation(self._se_object <= _get_se_object(other))


    def __lt__(self, other) -> 'Equation':
        return Equation(self._se_object < _get_se_object(other))


    def __ge__(self, other) -> 'Equation':
        return Equation(self._se_object >= _get_se_object(other))


    def __gt__(self, other) -> 'Equation':
        return Equation(self._se_object > _get_se_object(other))


class Equation(SymEngineObjectWrapper):
    """
    Represents a symbolic equation that can be converted into constraints for optimization.
    """

    def __init__(self, se_object):
        if isinstance(se_object, se_type.BooleanTrue): # Tautological Expression alway resolve to true
            raise TautologicalExpressionError()
        if isinstance(se_object, se_type.BooleanFalse):# Contradictory Expression alway resolve to false
            raise ContradictoryExpressionError()
        super().__init__(se_object)

    def generate_constraint(self, n_variable: int) -> Tuple[IndexValueArrayBuilder, NDArray]:
        """
        Generate a constraint mask and bounds from the symbolic equation.

        Returns
        -------
            - Constraint Mask
            - Constraint Bound

        Raise
        ------
            - ValueError: If n_variable is n <= 0.
            - ValueError: If the equation contains non-linear (quadratic or higher) terms.

        Warnings
        --------
            - Strict inequalities (e.g., <, >) are treated as non-strict (<=, >=).
        """
        if n_variable <= 0:
            raise ValueError("Cannot generate constraints object from equation for a non-positive number of variables")


        lhs, rhs = Expression(self._se_object.args[0]), Expression(self._se_object.args[1])
        eq = (rhs-lhs).expand()

        mask = IndexValueArrayBuilder((1, n_variable))
        const = 0

        for vars, coeffs in eq.terms().items():
            n_vars = len(vars)

            if n_vars == 0:
                const += coeffs
            elif n_vars == 1:
                mask.append(0, vars[0].problem_index(), coeffs)
            else:
                raise ValueError(
                    "Quadratic terms are not supported in constraints. "
                    "Please ensure that the expression contains only linear terms when using it as a constraint."
                )

        const = -const # inverse constant because it should be on the other side of the equation

        bounds = []
        if isinstance(self._se_object, se_type.Equality):
            bounds = [const, const]
        elif isinstance(self._se_object, se_type.LessThan):
            bounds = [const, np.nan]
        else: # strictly lesser than

            warnings.warn(
                "TitanQ does not support strictly less than (<) or strictly greater than (>) constraints."
                "These will be treated as less than or equal to (<=) or greater than or equal to (>=) constraints instead."
                "If this behavior is not desired, please revise the expression accordingly."
            )
            bounds = [const, np.nan]

        return mask, np.array(bounds, dtype=np.float32)


def _get_se_object(self: Union['SymEngineObjectWrapper', Any]) -> Union[se_type.Basic, Any]:
    """
    Return the underlying symengine object if the object is a SymengineObjectWrapper.
    Otherwise, return the object itself.

    Returns
    -------
        The unwrapped symengine object or the original object.
    """
    if isinstance(self, SymEngineObjectWrapper):
        return self._se_object
    else:
        return self


def _expand_term(term: se_type.Basic) -> ExpandedVariableList:
    """
    Decompose a mathematical term into its constituent variables.

    Returns
    -------
        A tuple containing the expanded variables.

    Raise
    ------
        NotImplementedError: If the term type is not supported.

    Example
    -------
        >>> expand_term(x * y)
        (x, y)
        >>> expand_term(x**3 * y**2)
        (x, x, x, y, y)
    """
    # Constant term
    if isinstance(term, (se_type.One, se_type.Zero)):
        return ExpandedVariableList()

    # Case for a single variable
    if isinstance(term, se_type.Symbol):
        return ExpandedVariableList(term)

    # Multiple variables of the same
    if isinstance(term, se_type.Pow):
        return ExpandedVariableList(*[term.base] * int(term.exp))

    # Multiple different variable
    if isinstance(term, se_type.Mul):
        return ExpandedVariableList(*(var for arg in term.args for var in _expand_term(arg)))

    raise NotImplementedError(f"Cannot split term into individual variable (type={type(term)})")


# Copyright (c) 2024, InfinityQ Technology, Inc.

import logging
from typing import Optional

from titanq._model.array.arraybuilder.array_builder_collection import ArrayBuilderCollection
from titanq._model.array.arraylike import ArrayLike

from .base_constraints import BaseConstraints
from ...errors import ConstraintSizeError


log = logging.getLogger("TitanQ")


class Constraints(BaseConstraints):
    def __init__(self) -> None:
        super().__init__()

        self._constraint_weights: ArrayBuilderCollection = ArrayBuilderCollection()
        self._constraint_bounds: ArrayBuilderCollection = ArrayBuilderCollection()


    def is_empty(self) -> bool:
        """Return if all constraints are empty."""
        return self._constraint_weights.is_empty() and self._constraint_bounds.is_empty()


    def add_constraint(self, num_variables: int, constraint_weights: ArrayLike, constraint_bounds: ArrayLike) -> None:
        """
        Add a constraint to the existing ones

        :param num_variables: the number of variables from the model
        :param constraint_weights: constraint_weights to append to the existing ones.
        :param constraint_bounds: constraint_bounds to append to the existing ones.

        :raises ConstraintSizeError:
            Number of constraints is different than the number of
            variables.
        """
        # shape validation
        if constraint_weights.shape()[1] != num_variables:
            raise ConstraintSizeError(
                "Constraint mask shape does not match the number of variables. "
                + f"Number of constraints: {constraint_weights.shape()[1]}, "
                + f"Number of variables: {num_variables}")


        self._constraint_weights.append(constraint_weights)
        self._constraint_bounds.append(constraint_bounds)


    def weights(self) -> Optional[ArrayLike]:
        """Returns the weights constraints."""
        if self._constraint_weights.is_empty():
            return None
        return self._array_like_factory.create_minimal(self._constraint_weights)


    def bounds(self) -> Optional[ArrayLike]:
        """Returns the bounds constraints."""
        if self._constraint_bounds.is_empty():
            return None
        return self._array_like_factory.create_numpy(self._constraint_bounds)



# Copyright (c) 2025, InfinityQ Technology, Inc.

"""
Milestone events are some type of events that would just occur at some point.
Nothing special about them, they just happen and something is instantly done.
"""

from typing_extensions import override

from titanq._event import Event
from titanq._event.sink import Sink


class UploadStartEvent(Event):
    """Milestone event when the SDK is starting to upload some files."""

    @override
    def output(self, sink: Sink) -> None:
        sink.display(
            sink.content_builder()
                .add_emoji(":outbox_tray:")
                .add_whitespace()
                .add_text("Starting data upload")
                .build()
        )


class GotComputationIDEvent(Event):
    """Milestone event when the SDK has obtained the computation ID from TitanQ"""

    def __init__(self, computation_id: str):
        self._computation_id = computation_id

    @override
    def output(self, sink: Sink):
        sink.display(
            sink.content_builder()
                .add_emoji(":id:")
                .add_whitespace()
                .add_text(f"Computation ID: {self._computation_id}")
                .build()
        )


class OptimizationCompletedEvent(Event):
    """Milestone event when the SDK has finished optimizing."""

    @override
    def output(self, sink: Sink) -> None:
        sink.display(
            sink.content_builder()
                .add_emoji(":white_check_mark:")
                .add_whitespace()
                .add_text("Optimization completed!")
                .build()
        )


class ComputationFailedEvent(Event):
    """Milestone event when the SDK has acknowledged that the computation has failed."""

    @override
    def output(self, sink: Sink) -> None:
        sink.display(
            sink.content_builder()
                .add_emoji(":exclamation:")
                .add_whitespace()
                .add_text("Computation failed")
                .build()
        )


# Copyright (c) 2025, InfinityQ Technology, Inc.

from datetime import datetime
from typing import List, Optional
import uuid
from pydantic import BaseModel, Field, RootModel, field_serializer


class ComputationHistoryRequest(BaseModel):
    id: Optional[uuid.UUID] = None
    limit: Optional[int] = None
    offset: Optional[int] = None
    start_date: Optional[datetime] = None
    end_date: Optional[datetime] = None

    @field_serializer("start_date", "end_date")
    def serialize_date(self, value: Optional[datetime]) -> Optional[str]:
        if value:
            return value.strftime("%Y-%m-%d")
        return None


class Status(BaseModel):
    status: str = Field(alias="Status")
    timestamp: datetime = Field(alias="TimeStamp")
    computation_id: uuid.UUID = Field(alias="ComputationID")
    note: str = Field(alias="Note")


class ComputationHistory(BaseModel):
    id: uuid.UUID
    account_id: uuid.UUID
    cost: int
    status: List[Status]


class ComputationHistoryResponse(RootModel):
    root: List[ComputationHistory]

    def __len__(self):
        return len(self.root)

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item):
        return self.root[item]


# Copyright (c) 2025, InfinityQ Technology, Inc.


import logging
from typing import List, Tuple
import numpy as np

log = logging.getLogger("TitanQ")


def row_sums_under_cardinality(sum, numpy_array: np.ndarray) -> Tuple[List[int], List[int]]:
    """
    Returns all rows indices that their values sum is equal/less than the given
    cardinality in the `numpy_array`.

    :param sum: sum of all values in each row
    :param numpy_array: array to compare to the sum

    :return: A tuple of the equal indices and less indices
    """
    equal_indices = np.where(sum == numpy_array)[0]
    less_indices = np.where(sum < numpy_array)[0]

    return (equal_indices, less_indices)


def compare_first_col_lower_second_col(mask: np.ndarray) -> bool:
    """Returns if the numpy array has every first column lower than the second."""
    if mask.ndim != 2 or mask.shape[1] != 2:
        raise ValueError("`compare_first_col_lower_second_col()` can only be used with arrays of shape (N, 2)")

    # filter first np.nan and np.inf
    valid_mask = np.isfinite(mask[:, 0]) & np.isfinite(mask[:, 1])
    return np.all(mask[valid_mask, 0] < mask[valid_mask, 1])


def are_values_unsigned_integer(mask: np.ndarray) -> bool:
    """Returns if the numpy array if every value is an unsigned integer"""
    if not np.issubdtype(mask.dtype, np.integer):
        return False # not integer array

    if np.any(mask < 0):
        return False # negative values found

    return True


# Copyright (c) 2024, InfinityQ Technology, Inc.
"""
``OptimizeResponse`` data object that is returned from the model when optimizing a problem.
"""

import logging
from typing import Any, Dict, List, Optional, Tuple, Union
import uuid
import warnings

import numpy as np

from titanq import errors

from .variables import Variable

log = logging.getLogger('TitanQ')

class OptimizeResponse:
    """
    Object containing the optimization results and all of the metrics returned by the solver.
    """
    def __init__(self, variables: List[Variable], result_array: np.ndarray, metrics: Dict[str, Any]) -> None:

        # extract all the result for each given variable
        # The result array is a 2d array where each line is a different result of the same problem
        result_by_variable: Dict[str, List[List[float]]] = {}
        for variable in variables:
            name = variable.parent_name()

            if name not in  result_by_variable:
                # if variable is not yet in the dict, create empty 2d list
                result_by_variable[name] = [[] for _ in range(len(result_array))]

            for result_index, full_result in enumerate(result_array):
                result_by_variable[name][result_index].append(full_result[variable.problem_index()])


        self._result_by_variable = {k: np.array(v, dtype=np.float32) for k, v in result_by_variable.items()}
        self._metrics = metrics
        self._all_results = result_array


    def __getattr__(self, attr: str):
        # if attribute is the name of a variable
        try:
            return self._result_by_variable[attr]
        except KeyError:
            pass

        # This is to keep compatibility with older version of SDK
        if attr == "ising_energy":
            try:
                return self.__getattr__("solutions_objective_value")
            except (AttributeError, KeyError):
                pass


        warnings.warn(
            'Obtaining metrics directly as an attribute is deprecated. Use computation_metrics() or original_input_params() instead.',
            DeprecationWarning,
            stacklevel=2
        )


        # check inside computation metrics and original params for the attribute
        try:
            return self.computation_metrics(attr)
        except KeyError:
            pass

        try:
            return self.original_input_params(attr)
        except KeyError:
            pass

        # was not found, try the older behavior
        try:
            return self._metrics[attr]
        except KeyError:
            raise AttributeError(attr)


    def result_vector(self) -> np.ndarray:
        """
        The result vector

        Returns
        -------
        The result vector of this optimization.
        """
        return self._all_results


    def result_items(self) -> List[Tuple[float, np.ndarray]]:
        """
        ex. [(-10000, [0, 1, 1, 0]), (-20000, [1, 0, 1, 0]), ...]

        Returns
        -------
        List of tuples containing the solutions objective value and its corresponding result vector
        """

        solutions_objective_value = self.ising_energy
        return [(solutions_objective_value[i], self._all_results[i]) for i in range(len(self._all_results))]


    def computation_id(self) -> uuid.UUID:
        """
        The computation id is a Universal unique id that identify this computation inside the TitanQ platform.

        Provide this id on any support request to the InfinityQ.

        Returns
        -------
        The computation id of this solve.

        Raises
        ------
        UnexpectedServerResponseError
            The server response doesn't contain a computation id
        """
        try:
            computation_id = self._metrics['computation_id']
        except KeyError as ex:
            raise errors.UnexpectedServerResponseError(f"Failed to fetch '{ex}' from the metrics response file, please contact InfinityQ support for more help")

        return computation_id


    def computation_metrics(self, key: str = None) -> Any:
        """
        The computation metrics the solver returns

        Returns
        -------
        All computation metrics if no key is given of the specific metrics with the associated key if one is provided.

        Raises
        ------
        UnexpectedServerResponseError
            The server response doesn't contain the computation metrics or the passed key
        """
        try:
            metrics = self._metrics['computation_metrics']
            if key:
                metrics = metrics[key]
        except KeyError as ex:
            raise errors.UnexpectedServerResponseError(f"Failed to fetch '{ex}' from the metrics response file, please contact" \
                f" InfinityQ support for more help and provide the following computation ID: {self.computation_id()}") from ex

        return metrics


    def parameters_used(self, key: str = None) -> Any:
        """
        The actual parameters used by the solver. If a parameter was not used, the value will be set to None.

        Returns
        -------
        The parameters used by the solver

        Raises
        ------
        UnexpectedServerResponseError
            The server response doesn't contain the parameters used or the passed key
        """
        try:
            metrics = self._metrics['parameters_used']
            if key:
                metrics = metrics[key]
        except KeyError as ex:
            raise errors.UnexpectedServerResponseError(f"Failed to fetch '{ex}' from the metrics response file, please contact" \
                f" InfinityQ support for more help and provide the following computation ID: {self.computation_id()}") from ex

        return metrics


    def problem_shape(self, key: str = None) -> Any:
        """
        The problem shape analyzed by the solver

        Returns
        -------
        The problem shape

        Raises
        ------
        UnexpectedServerResponseError
            The server response doesn't contain the problem shape or the passed key
        """
        try:
            metrics = self._metrics['problem_shape']
            if key:
                metrics = metrics[key]
        except KeyError as ex:
            raise errors.UnexpectedServerResponseError(f"Failed to fetch '{ex}' from the metrics response file, please contact" \
                f" InfinityQ support for more help and provide the following computation ID: {self.computation_id()}") from ex

        return metrics


    def solver_used(self) -> str:
        """
        The solver class used for the computation

        Returns
        -------
        The solver class used

        Raises
        ------
        UnexpectedServerResponseError
            The server response doesn't contain the solver used
        """
        try:
            solver_used = self._metrics['solver_used']
        except KeyError as ex:
            raise errors.UnexpectedServerResponseError(f"Failed to fetch '{ex}' from the metrics response file, please contact" \
                f" InfinityQ support for more help and provide the following computation ID: {self.computation_id()}") from ex

        return solver_used


    def constraint_violations(self) -> Tuple[Optional[List[int]], Optional[List[List[int]]]]:
        """
        Obtain constraint violations for each engine (``num_engines``).
        If no constraints were used in the computation, the values will be None.

        Return
        ------
        A tuple of two lists
        - First list indicate the total number of constraint violations for each engine
        - Second list contains a list of the constraint violation indexes for each engine

        """
        return (
            self.computation_metrics("total_constraint_violations"),
            self.computation_metrics("constraint_violation_indices")
        )


    def total_number_of_constraints(self) -> int:
        """
        Returns the number of constraints passed to the solver for the computation.

        Returns
        -------
        The number of constraints, or 0 if no constraints were defined in the request.
        """
        constraint_weights_shape: Tuple[int, int] = self.original_input_params("constraint_weights_shape")

        return 0 if constraint_weights_shape is None else constraint_weights_shape[0]


    def original_input_params(self, key: str = None) -> Any:
        """
        .. deprecated:: 0.29.0
            Use parameters_used() instead.

        The original input params sent to the solver

        Returns
        -------
        All original params if no key is given. A specific param with the associated key if one is provided.

        Raises
        ------
        UnexpectedServerResponseError
            The server response doesn't contain the original input parameters or the passed key
        """
        try:
            metrics = self._metrics['original_input_params']
            if key:
                metrics = metrics[key]
        except KeyError as ex:
            raise errors.UnexpectedServerResponseError(f"Failed to fetch '{ex}' from the metrics response file, please contact" \
                f" InfinityQ support for more help and provide the following computation ID: {self.computation_id()}") from ex

        return metrics


    def metrics(self, key: str = None) -> Union[str, Dict[str, Any]]:
        """
        .. deprecated:: 0.7.0
            Use computation_metrics() or original_input_params() instead.

        Returns
        -------
        All metrics if no key is given. A specific metric with the associated key if one is provided.

        Raises
        ------
        UnexpectedServerResponseError
            The server response doesn't contain the metrics or the passed key
        """
        warnings.warn(
            'Calling metrics() is deprecated. Use computation_metrics() or original_input_params() instead.',
            DeprecationWarning,
            stacklevel=2
        )
        if key:
            try:
                metrics = metrics[key]
            except KeyError as ex:
                raise errors.UnexpectedServerResponseError(f"Failed to fetch '{ex}' from the solver response, please contact" \
                f" InfinityQ support for more help and provide the following computation ID: {self.computation_id()}") from ex

            return metrics
        else:
            return self._metrics


# Copyright (c) 2024, InfinityQ Technology, Inc.

####################################
# Errors specific to the TitanQ SDK.
####################################

from contextlib import contextmanager
from typing import Optional


class TitanqError(Exception):
    """
    Base TitanQ error, optionally can be instantiated with a computation id
    """

    def __init__(self, message: str = "", computation_id: Optional[str] = None):
        self.original_message = message
        self.computation_id = computation_id
        super().__init__(self._compose_message())

    def with_computation_id(self, computation_id: str) -> 'TitanqError':
        return type(self)(self.original_message, computation_id=computation_id)

    def _compose_message(self):
        if self.computation_id:
            return f"computation ID: {self.computation_id}: {self.original_message}"
        return self.original_message

    def __str__(self):
        return self._compose_message()


@contextmanager
def titanq_error_including_computation_id(id: str):
    """Computation manager that will include a 'TitanqError' with a computation id"""
    try:
        yield
    except TitanqError as e:
        raise e.with_computation_id(id) from e


#######################
# SERVER RELATED ERRORS
#######################
class ClientError(TitanqError):
    """Base client-side http error"""

class NotEnoughCreditsError(ClientError):
    """Not enough credits left"""

    def __init__(self, message="Not enough credits left", *args, **kwargs):
        super().__init__(message, *args, **kwargs)

class BadRequest(ClientError):
    """The request sent to TitanQ is not valid"""

class UnexpectedServerResponseError(TitanqError):
    """Response from the TitanQ server is not as expected"""

class MissingFinishedStatusError(TitanqError):
    """The computation does not include a finished status"""

class ConnectionError(TitanqError):
    """Error due to a connection issue with an external resource"""

class ServerError(TitanqError):
    """Unexpected condition prevented the TitanQ server to fulfill the request"""

class EmptyResultError(TitanqError):
    """Results were downloaded but they are empty"""

class UnsolvableRequestError(TitanqError):
    """TitanQ cannot solve this combination of parameters"""

    def __init__(self, message="TitanQ cannot solve this combination of parameters", *args, **kwargs):
        super().__init__(message, *args, **kwargs)

class ComputationFailedError(TitanqError):
    """The computation failed"""

class UnknownError(TitanqError):
    """An error happened in TitanQ, but the SDK is unable to know what is the source of the problem."""
    def __init__(self, message="TitanQ has experienced an unknown error, please contact (support@infinityq.tech).", *args, **kwargs):
        super().__init__(message, *args, **kwargs)


######################
# MODEL RELATED ERRORS
######################
class MissingTitanqApiKey(TitanqError):
    """TitanQ Api key is missing"""

class MissingVariableError(TitanqError):
    """Variable has not already been registered"""

class VariableAlreadyExist(TitanqError):
    """Variable with the same name already exist"""

class MissingObjectiveError(TitanqError):
    """Objective has not already been registered"""

class ConstraintSizeError(TitanqError):
    """Unexpected number of constraints"""

class ConstraintAlreadySetError(TitanqError):
    """A constraint has already been set"""

class ObjectiveAlreadySetError(TitanqError):
    """An objective has already been set"""

class TautologicalExpressionError(TitanqError):
    """
    Exception raised when an expression is tautological (always true).

    This exception indicates that the provided expression is redundant
    and does not add meaningful constraints or information.
    """
    def __init__(self, message="The provided expression is tautological and always evaluates to True, regardless of the variable values.", *args, **kwargs):
        super().__init__(message, *args, **kwargs)

class ContradictoryExpressionError(TitanqError):
    """
    Exception raised when an expression is contradictory (always false).

    This exception indicates that the provided expression is invalid
    as it represents an impossible condition.
    """
    def __init__(self, message="The provided expression is contradictory and always evaluates to False, regardless of the variable values.", *args, **kwargs):
        super().__init__(message, *args, **kwargs)


####################
# MPS RELATED ERRORS
####################
class MpsParsingError(TitanqError):
    """Base class for any error related to the MPS files parsing module"""

class MpsConfiguredModelError(MpsParsingError):
    """Passed model is already configured"""

class MpsMissingValueError(MpsParsingError):
    """A required value is missing"""

class MpsMissingSectionError(MpsParsingError):
    """A required section is missing"""

class MpsMalformedFileError(MpsParsingError):
    """The file is malformed"""

class MpsUnexpectedValueError(MpsParsingError):
    """Found an unexpected value"""

class MpsUnsupportedError(MpsParsingError):
    """Found an unsupported value"""



# Copyright (c) 2025, InfinityQ Technology, Inc.

def generate_word_animation(word: str):
    """Generates a cool TITANQ ASCII animation, returned in a list by frames"""
    word_len = len(word)
    frame = "." * word_len
    frames = [frame]

    for (target_position, letter) in enumerate(word):
        current_index = word_len - 1
        frame = frame[:-1] + letter
        frames.append(frame)

        while current_index != target_position-1:
            frame = (frame[:current_index] + letter).ljust(word_len, ".")
            frames.append(frame)
            current_index -= 1

    frames.append("TITANQ")
    frames.append("TITAN•")
    frames.append("TITA•Q")
    frames.append("TIT•NQ")
    frames.append("TI•ANQ")
    frames.append("T•TANQ")
    frames.append("•ITANQ")
    frames.append("••••••")
    frames.append("TITANQ")
    frames.append("••••••")
    frames.append("TITANQ")
    return frames


# Copyright (c) 2024, InfinityQ Technology, Inc.
from dataclasses import dataclass
from typing import Optional

from titanq.errors import MpsMalformedFileError


@dataclass
class MPSParseOptions:

    skip_empty_lines: Optional[bool] = False
    """Will skip empty lines found inside the .mps file"""

    def empty_line_check(self, index: int, content: str) -> bool:
        """
        return a boolean value if it should skip the line or not.
        If skip_empty_lines is not raised, it will raise an error instead

        :param index: line index we are checking to skip or not
        :param conent: content of the line

        :raises: MPSMalormedFile if the line is empty and the flag is not raised
        """
        if not self.skip_empty_lines and not content:
            raise MpsMalformedFileError(
                f"Line number '{index}' in the .mps file is empty, use 'skip_empty_lines' as an option to ignore"
            )
        return self.skip_empty_lines and not content # skip only if flag set and it's empty


# Copyright (c) 2024, InfinityQ Technology, Inc.

from abc import abstractmethod, ABC

from titanq._model.array.factory import ArrayLikeFactory


class BaseConstraints(ABC):

    def __init__(self):
        self._array_like_factory = ArrayLikeFactory()

    @abstractmethod
    def is_empty(self):
        """Return if all constraints are empty."""
        pass


# Copyright (c) 2025, InfinityQ Technology, Inc.

import io
from typing import Iterable, List, Tuple, Union

import numpy as np
from scipy.sparse import coo_array, save_npz

from titanq._model.array.arraybuilder import IndexKey, RealNumber
from titanq._model.array.arraylike import ArrayLike, util


class ScipyCooArray(ArrayLike):

    def __init__(self, array: coo_array):
        """
        Creates a ScipyCooArray implementing ArrayLike

        :param array: The scipy COOrdinate array itself
        """
        self._array = array

    def shape(self) -> Tuple[int, int]:
        return self._array.shape

    def density(self) -> float:
        if self._array.ndim == 1:
            # non-zero elements divided by length of the 1D array
            return self._array.nnz / self._array.shape[0]
        else:
            # non-zero elements divided by the size of the 2D array
            return self._array.nnz / (self._array.shape[0] * self._array.shape[1])

    def non_null_items(self) -> Iterable[Tuple[IndexKey, RealNumber]]:
        if self.ndim() == 1:
            for col, data in zip(self._array.col, self._array.data):
                yield (0, col), data
        else:
            for row, col, data in zip(self._array.row, self._array.col, self._array.data):
                yield (row, col), data

    def inner(self) -> coo_array:
        return self._array

    def ndim(self) -> int:
        return self._array.ndim

    def data_type(self) -> np.dtype:
        return self._array.dtype

    def is_binary(self):
        return np.all((self._array.data == 1) | (self._array.data == 0))

    def sum(self, axis=None) -> Union[int, np.ndarray]:
        return self._array.sum(axis=axis)

    def to_bytes(self) -> bytes:
        buffer = io.BytesIO()
        save_npz(buffer, self._array)
        return buffer.getvalue()

    def reshape_to_2d(self) -> None:
        if self.ndim() != 1:
            raise ValueError("Input array is not 1-D. Cannot reshape array.")

        self._array = self._array.reshape(1, -1)

    def repeat_rows(self) -> None:
        self._array = coo_array(np.repeat(self._array.toarray(), 2).reshape(-1, 2))

    def isnan(self) -> bool:
        return np.isnan(self._array.data).any()

    def isinf(self) -> bool:
        return np.isinf(self._array.data).any()

    def is_first_col_lower_second_col(self) -> bool:
        return util.compare_first_col_lower_second_col(self._array.toarray())

    def are_values_unsigned_integer(self) -> bool:
        return util.are_values_unsigned_integer(self._array.toarray())

    def row_sums_under_cardinality(self, numpy_array: np.ndarray) -> List[int]:
        return util.row_sums_under_cardinality(self.sum(axis=1), numpy_array)


# Copyright (c) 2024, InfinityQ Technology, Inc.

import logging
from typing import Optional

from titanq._model.array.arraybuilder.array_builder_collection import ArrayBuilderCollection
from titanq._model.array.arraylike import ArrayLike

from .base_constraints import BaseConstraints
from ...errors import ConstraintSizeError


log = logging.getLogger("TitanQ")


class QuadConstraints(BaseConstraints):
    def __init__(self) -> None:
        super().__init__()

        self._quad_constraint_weights: ArrayBuilderCollection = ArrayBuilderCollection()
        self._quad_constraint_bounds: ArrayBuilderCollection = ArrayBuilderCollection()
        self._quad_constraint_linear_weights: ArrayBuilderCollection = ArrayBuilderCollection()

    def is_empty(self) -> bool:
        """return if all constraints are empty"""

        return (
            self._quad_constraint_weights.is_empty() and
            self._quad_constraint_bounds.is_empty() and
            self._quad_constraint_linear_weights.is_empty()
        )


    def add_constraint(
        self,
        num_variables: int,
        quad_constraint_weights: ArrayLike,
        quad_constraint_bounds: ArrayLike,
        quad_constraint_linear_weights: Optional[ArrayLike] = None
    ) -> None:
        """
        Add a quadratic constraint to the existing ones

        :param num_variables: the number of variables from the model
        :param quad_constraint_weights: quadratic constraint weights to append to the existing ones.
        :param quad_constraint_bounds: quadratic constraint bounds to append to the existing ones.
        :param quad_constraint_linear_weights: quadratic constraint linear weights to append to the existing ones.

        :raises ConstraintSizeError:
            Number of constraints is different than the number of
            variables.
        """
        # shape validation
        if quad_constraint_weights.shape() != (num_variables, num_variables):
            raise ConstraintSizeError(
                "Invalid constraint_mask shape: expected (N,N) where N is the number of variables "
                + f"({num_variables},{num_variables}), but got {quad_constraint_weights.shape()}.")

        if quad_constraint_linear_weights is not None and quad_constraint_linear_weights.shape() != (1, num_variables):
                raise ValueError(
                    "Invalid constraint_linear_weights shape: expected (N,) where N is "
                    + f"the number of variables ({num_variables}), "
                    + f"but got {quad_constraint_linear_weights.shape()}.")

        self._quad_constraint_weights.append(quad_constraint_weights)
        self._quad_constraint_bounds.append(quad_constraint_bounds)

        if quad_constraint_linear_weights is not None:
            self._quad_constraint_linear_weights.append(quad_constraint_linear_weights)


    def weights(self) -> Optional[ArrayLike]:
        """Returns the quadratic weights constraints."""
        if self._quad_constraint_weights.is_empty():
            return None
        return self._array_like_factory.create_minimal(self._quad_constraint_weights)


    def bounds(self) -> Optional[ArrayLike]:
        """Returns the quadratic bounds constraints."""
        if self._quad_constraint_bounds.is_empty():
            return None
        return self._array_like_factory.create_numpy(self._quad_constraint_bounds)

    def linear_weights(self) -> Optional[ArrayLike]:
        """Return the quadratic linear weights constraints."""
        if self._quad_constraint_linear_weights is None:
            return None
        if self._quad_constraint_linear_weights.is_empty():
            return None
        return self._array_like_factory.create_minimal(self._quad_constraint_linear_weights)



# Copyright (c) 2025, InfinityQ Technology, Inc.

from datetime import datetime
from enum import Enum
from typing import List, Optional
from typing_extensions import override, Self

from pyfiglet import Figlet
from rich.columns import Columns
from rich.console import Group, RenderableType
from rich.live import Live
from rich.panel import Panel
from rich.progress import Progress, TextColumn, TimeElapsedColumn, BarColumn, TaskProgressColumn
from rich.rule import Rule
from rich.spinner import Spinner, SPINNERS
from rich.table import Table
from rich.text import Text
from rich.tree import Tree

from titanq._event import Sink
from titanq._event.sink import ContentBuilder, DefinedProgress, UndefinedProgress
from titanq._event.sink.display_util import generate_word_animation
from titanq._util.decorators import chainable


# append rich library spinners our titanq animation
SPINNERS["titanq"] = {
    "interval": 120,
    "frames": generate_word_animation("TITANQ")
}


class PrettyTextRenderableTag(Enum):
    """
    Enums for any 'PrettyTextRenderable' tag.
    Needed to distinguish a renderable from another.
    """
    HEADER_CONTENT = 1
    UNDEFINED_PROGRESS_CONTENT = 2
    DEFINED_PROGRESS_CONTENT = 3
    TEXT_CONTENT = 4


class PrettyTextRenderable:
    """Wrapper arround rich 'RenderableType' that also contains a tag."""

    def __init__(self, content: RenderableType, tag: PrettyTextRenderableTag):
        self._content = content
        self._tag = tag

    def content(self) -> RenderableType:
        """The content of the renderable"""
        return self._content

    def tag(self) -> PrettyTextRenderableTag:
        """The tag of the renderable"""
        return self._tag


class PrettyTextContentBuilder(ContentBuilder[PrettyTextRenderable]):

    def __init__(self):
        self._content: Text = Text()
        self._color: Optional[str] = None
        self._spinner: Spinner = None

    @override
    def build(self) -> PrettyTextRenderable:
        # color the content only if set
        if self._color:
            self._content.stylize_before(self._color)

        if self._spinner:
            table = Table.grid(padding=(0, 1))
            table.add_row(self._content, self._spinner)
            return PrettyTextRenderable(table, PrettyTextRenderableTag.TEXT_CONTENT)

        return PrettyTextRenderable(self._content, PrettyTextRenderableTag.TEXT_CONTENT)

    @override
    @chainable
    def set_color(self, color: str) -> Self:
        self._color = color

    @override
    @chainable
    def add_emoji(self, emoji: str) -> Self:
        emoji = Text.from_markup(emoji)
        self._content.append(emoji)

    @override
    @chainable
    def add_text(self, text: str) -> Self:
        self._content.append(text)

    @override
    @chainable
    def add_whitespace(self) -> Self:
        self._content.append(" ")

    @override
    @chainable
    def add_tailing_spinner(self) -> Self:
        self._spinner = Spinner("titanq", style="magenta")


class PrettyTextDefinedProgress(DefinedProgress[PrettyTextRenderable]):

    def __init__(self):
        self._progress = Progress(
            TextColumn("{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            TimeElapsedColumn(),
        )

        # needed to keep track which update is it for
        self._progress_tasks: List[int] = []

    @override
    def add_task(self, renderable: PrettyTextRenderable, total: int) -> int:
        return self._progress.add_task(renderable.content(), total=total, time=datetime.now().strftime("[%H:%M:%S]"))

    @override
    def content(self) -> PrettyTextRenderable:
        return PrettyTextRenderable(self._progress, PrettyTextRenderableTag.DEFINED_PROGRESS_CONTENT)

    @override
    def update(self, handle: int, new_value: int, renderable: Optional[PrettyTextRenderable] = None) -> None:
        if renderable is not None:
            self._progress.update(handle, completed=new_value, description=renderable.content())
        else:
            self._progress.update(handle, completed=new_value)

    @override
    def finalize(self) -> None:
        # stop all tasks, whatever the status
        for task in self._progress_tasks:
            self._progress.stop_task(task)

        # stop the progress bar
        self._progress.stop()


class PrettyTextUndefinedProgress(UndefinedProgress[PrettyTextRenderable]):

    @override
    def update(self, renderable: PrettyTextRenderable, additional_renderables: Optional[List[PrettyTextRenderable]] = None) -> PrettyTextRenderable:
        grid = Table.grid(padding=(0, 1))
        tree = Tree(label=renderable.content(), guide_style="dim")
        grid.add_row(tree)

        if additional_renderables:
            for renderable in additional_renderables:
                tree.add(renderable.content())

        return PrettyTextRenderable(grid, PrettyTextRenderableTag.UNDEFINED_PROGRESS_CONTENT)


class DisplayHandler:

    def __init__(self):
        self._live = Live(refresh_per_second=10)
        self._live_entries: Group = Group()

        self._previous_tag: PrettyTextRenderableTag = None

    def start(self) -> None:
        """Start the display handler."""
        self._live.start()

    def stop(self) -> None:
        """Stops the display handler."""
        # ensure that last frame is sent (especially for Jupyter as sometime the last frame is not sent)
        self._live.refresh()
        self._live.stop()

    def update(self, entry: RenderableType, tag: PrettyTextRenderableTag) -> None:
        """Update the display content."""
        # case where we want to update the last undefined progress content,
        # instead of adding a new entry.
        if self._is_previous_and_current_tag_undefined_progress(tag):
            self._update_last(entry)
        # when creating a defined progress renderable, a renderable is updated multiple times for the same entry.
        elif self._is_previous_and_current_tag_defined_progress(tag):
            self._update_last(entry)
        else:
            self._add_new(entry)

        # update with newest tag
        self._previous_tag = tag

    def _is_previous_and_current_tag_undefined_progress(self, tag) -> bool:
        return (
            self._previous_tag is PrettyTextRenderableTag.UNDEFINED_PROGRESS_CONTENT and
            tag is PrettyTextRenderableTag.UNDEFINED_PROGRESS_CONTENT
        )

    def _is_previous_and_current_tag_defined_progress(self, tag: PrettyTextRenderableTag):
        return (
            self._previous_tag is PrettyTextRenderableTag.DEFINED_PROGRESS_CONTENT and
            tag is PrettyTextRenderableTag.DEFINED_PROGRESS_CONTENT)

    def _add_new(self, entry: RenderableType) -> None:
        """Adds a new entry to the display"""
        self._live_entries.renderables.append(entry)
        self._live.update(self._live_entries)

    def _update_last(self, entry: RenderableType) -> None:
        """
        Updates the last entry of the display.
        Meaning it will remove it and set the new one
        """
        self._live_entries.renderables.pop()
        self._add_new(entry)


class PrettyTextSink(Sink[PrettyTextRenderable]):
    """
    The PrettyTextSink utilizes rich library to enhance the output and make it
    as beautiful as possible and pleasing for the end-users.

    It offers features such as Emoji's support, colors and some progress bars.
    """

    def __init__(self,
        version: str,
        show_header: bool,
        start_separator: Optional[str] = None
    ):
        """
        Initialize the pretty sink logger.

        :param version: Version shown in the sink
        :param show_header: If True, will show a header in the startup
        :param start_separator: Separator text to include if set that will be shown in the startup
        """
        self._version = version
        self._show_header = show_header
        self._start_separator = start_separator

        self._display_handler = DisplayHandler()

    @override
    def start(self) -> None:
        self._display_handler.start()
        self._start_display()

    @override
    def stop(self) -> None:
        self._display_handler.stop()

    @override
    def display(self, renderable: PrettyTextRenderable, with_timestamp: Optional[bool] = True) -> None:
        timestamp = Text(datetime.now().strftime("[%H:%M:%S]"), style="dim #39c5cf")

        columns = Columns([renderable.content()])
        if with_timestamp:
            columns.renderables.insert(0, timestamp)

        self._display_handler.update(columns, renderable.tag())

    @override
    def content_builder(self) -> ContentBuilder[PrettyTextRenderable]:
        return PrettyTextContentBuilder()

    @override
    def defined_progress(self) -> DefinedProgress[PrettyTextRenderable]:
        return PrettyTextDefinedProgress()

    @override
    def undefined_progress(self) -> UndefinedProgress[PrettyTextRenderable]:
        return PrettyTextUndefinedProgress()

    def _start_display(self) -> None:
        """What will be displayed when starting up"""
        if self._show_header:
            self.display(self._banner(version=self._version), with_timestamp=False)

        # display a separator that would indicate what optimization we are at
        if self._start_separator is not None:
            self._display_handler.update(
                Rule(self._start_separator, style=None),
                PrettyTextRenderableTag.TEXT_CONTENT
            )

    def _banner(self, version: str) -> PrettyTextRenderable:
        """
        Displays a banner.
        """
        #╭───────────────────── vX.Y.Z ──────────────────────╮
        #│      _________________              _______       │
        #│      ___  __/__(_)_  /______ _________  __ \      │
        #│      __  /  __  /_  __/  __ `/_  __ \  / / /      │
        #│      _  /   _  / / /_ / /_/ /_  / / / /_/ /       │
        #│      /_/    /_/  \__/ \__,_/ /_/ /_/\___\_\       │
        #│                                                   │
        #│                                                   │
        #╰───────────────────────────────────────────────────╯
        panel = Panel.fit(
            Figlet(font='speed').renderText("TitanQ"), style="purple",
            title=f"[italic bold purple]{version}[/]",
            padding=(0, 6),  # (top_bottom, left_right)
            border_style="purple",
        )
        return PrettyTextRenderable(panel, "header")



# Copyright (c) 2025, InfinityQ Technology, Inc.

from io import BytesIO
from typing import Annotated, Optional, Union

from pydantic import AfterValidator, BaseModel, ConfigDict, Field, SerializeAsAny, field_serializer, model_serializer

from titanq._api.model._util import UrlTruncatedOnDisplay


# FileName is either a Str or an URL.
# It Has a validation that will cast it into an URL if needed
# Meaning a simple string can alway be use when setting it
FileName = Annotated[
    Union[str, UrlTruncatedOnDisplay],
    AfterValidator(lambda s: UrlTruncatedOnDisplay(s) if s.startswith("http") else s),
]
InputFile = SerializeAsAny[Union[FileName, 'LocalFile']]


class LocalFile(BaseModel):
    """ A local file for numpy arrays """
    remote_name: FileName
    data: BytesIO = Field(exclude=True)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_serializer
    def serialize_model_as_str(self) -> str:
        if isinstance(self.remote_name, UrlTruncatedOnDisplay):
            return self.remote_name.non_obfuscated_value()
        return self.remote_name


class InputFiles(BaseModel):
    """ Input files for any location type """
    bias_file_name: InputFile
    variable_bounds_file_name: InputFile
    weights_file_name: Optional[InputFile] = None
    constraint_weights_file_name: Optional[InputFile] = None
    constraint_bounds_file_name: Optional[InputFile] = None
    quad_constraint_weights_file_name: Optional[InputFile] = None
    quad_constraint_bounds_file_name: Optional[InputFile] = None
    quad_constraint_linear_weights_file_name: Optional[InputFile] = None

    # We need to explicit the serialization because how Union[] work with pydantic
    @field_serializer("*")
    def serialize_input_file(self, value):
        if isinstance(value, UrlTruncatedOnDisplay):
            return value.non_obfuscated_value()
        return value


# Copyright (c) 2025, InfinityQ Technology, Inc.

"""
The Array module defines the structure for array management within the TitanQ SDK.

`ArrayBuilder` is an interface that enables its subclasses to construct arrays for use with the TitanQ SDK.

`ArrayLike` is an interface for objects that are considered "array-like." It provides a set of methods that
can be implemented by any class—whether it's a numpy.ndarray or a specialized type (such as scipy's coo_array)
to ensure compatibility with the expected behaviors of the SDK.

`ArrayLike` instances are created through the `ArrayLikeFactory`, ensuring consistency and simplicity across the SDK.

`Array` is an alias for the raw data arrays supported by TitanQ, meaning the array module can directly interact with
these types of arrays.
"""

from typing import Union

import numpy as np
from scipy.sparse import coo_array, csr_array


# Alias for the supported array types in the TitanQ SDK
Array = Union[np.ndarray, coo_array, csr_array]

# dtype to use for any numpy array like
_TITANQ_DTYPE = np.float32


# Copyright (c) 2025, InfinityQ Technology, Inc.

"""
Some decorators used across the module.
"""

from functools import wraps
from typing import Callable, TypeVar


T = TypeVar("T")


def chainable(method: Callable[[T], None]) -> Callable[[T], T]:
    """
    Decorator that would return self when used with a method.
    This is useful when you want to make chained methods.

    Usage:
        data = some_method().another_mother().even_another_method()
    """
    @wraps(method)
    def wrapper(self: T, *args, **kwargs) -> T:
        method(self, *args, **kwargs)
        return self
    return wrapper



# Copyright (c) 2024, InfinityQ Technology, Inc.
from typing import List, Optional
from warnings import warn

from titanq.errors import MpsMissingValueError
from .utils import BoundsType, ColumnsType, RowsType
from ._visitor import LineSection, Visitor


class MPSName(LineSection):
    def __init__(self, name: str) -> None:
        self._name = name

    def name(self) -> str: return self._name

    def _accept(self, visitor: Visitor):
        visitor._visit_name(self)

    def __str__(self):
        """.mps format"""
        return self._name


class MPSRow(LineSection):
    """MPS Row section handler"""
    def __init__(self, sense: RowsType, identifier: str) -> None:
        self._sense = sense
        self._identifier = identifier

    def sense(self) -> RowsType: return self._sense
    def identifier(self) -> str: return self._identifier

    def _accept(self, visitor: Visitor):
        if self._sense == RowsType.FREE_ROW:
            visitor._visit_free_row(self)
        else:
            visitor._visit_rows(self)

    def __str__(self):
        """.mps format"""
        return f"{self._sense}\t{self._identifier}"

class MPSColumns(LineSection):
    def __init__(self, identifier: str,row_identifier: str, coeff: float, type: ColumnsType) -> None:
        self._identifier = identifier
        self._row_identifier = row_identifier
        self._coeff = coeff
        self._type = type

    def identifier(self) -> str: return self._identifier
    def row_identifier(self) -> str: return self._row_identifier
    def coeff(self) -> float: return self._coeff
    def type(self) -> ColumnsType: return self._type

    def _accept(self, visitor: Visitor):
        visitor._visit_columns(self)

    def __str__(self):
        """.mps format"""
        return f"\t{self._identifier}\t{self._row_identifier}\t{self._coeff}"


class MPSRhs(LineSection):
    def __init__(self, identifier: str, row_identifier: str, coeff: float) -> None:
        self._identifier = identifier
        self._row_identifier = row_identifier
        self._coeff = coeff

    def identifier(self) -> str: return self._identifier
    def row_identifier(self) -> str: return self._row_identifier
    def coeff(self) -> float: return self._coeff

    def _accept(self, visitor: Visitor):
        visitor._visit_rhs(self)

    def __str__(self):
        return f"\t{self._identifier}\t{self._row_identifier}\t{self._coeff}"

class MPSRanges(LineSection):
    def __init__(self, identifier: str, row_identifier: str, coeff: float) -> None:
        self._identifier = identifier
        self._row_identifier = row_identifier
        self._coeff = coeff

    def identifier(self) -> str: return self._identifier
    def row_identifier(self) -> str: return self._row_identifier
    def coeff(self) -> float: return self._coeff

    def _accept(self, visitor: Visitor):
        visitor._visit_ranges(self)

    def __str__(self):
        """.mps format"""
        return f"\t{self._identifier}\t{self._row_identifier}\t{self._coeff}"


class MPSBounds(LineSection):
    def __init__(self, identifier: str, type: BoundsType, column_identifier: str, value: Optional[float] = None) -> None:
        self._identifier = identifier
        self._type = type
        self._column_identifier = column_identifier
        self._value = value

        # check for mandatory values in some type of bounds
        if self._type in BoundsType.needs_value():
            if self._value is None:
                raise MpsMissingValueError(
                    f"No value was assigned for a bound of type '{self._type}', "
                    f"with the identifier '{self._identifier}'")
        else:
            # a value assigned to a type that does not require one
            if self._value is not None:
                warn(f"A value was assigned for a bound of type '{self._type}', with the identifier '{self._identifier}', "
                      "but this one does not require any value, the value will be ignored instead")

    def identifier(self) -> str: return self._identifier
    def type(self) -> BoundsType: return self._type
    def column_identifier(self) -> str: return self._column_identifier
    def value(self) -> Optional[float]: return self._value

    def _accept(self, visitor: Visitor):
        visitor._visit_bounds(self)

    def __str__(self):
        """.mps format"""
        return f"{self._type}\t{self._identifier}\t{self._column_identifier}\t{self._value}"


class MPSQuadobj(LineSection):
    def __init__(self, row_identifier: str, column_identifier: str, value: float) -> None:
        self._row_identifier = row_identifier
        self._column_identifier = column_identifier
        self._value = value

    def row_identifier(self) -> str: return self._row_identifier
    def column_identifier(self) -> str: return self._column_identifier
    def value(self) -> float: return self._value

    def _accept(self, visitor: Visitor):
        visitor._visit_quadobj(self)

    def __str__(self):
        """.mps format"""
        return f"{self._row_identifier}\t{self._column_identifier}\t{self._value}"

class MPSEndata(LineSection):
    def _accept(self, visitor: Visitor):
        visitor._visit_endata()


# Main MPS object containing all info
class MPSObject(Visitor):

    def __init__(self) -> None:
        self.name: str = None
        self.free_row: MPSRow = None # called as free row, which is the objective row
        self.objective_offset: Optional[MPSRhs] = None
        self.rows: List[MPSRow] = []
        self.columns: List[MPSColumns] = []
        self.rhs: List[MPSRhs] = []
        self.ranges: List[MPSRanges] = []
        self.bounds: List[MPSBounds] = []
        self.quadobj: List[MPSQuadobj] = []
        self.endata: bool = False

    # visitor methods
    def _visit_name(self, name: MPSName): self.name = name.name()
    def _visit_rows(self, row: MPSRow): self.rows.append(row)
    def _visit_columns(self, column: MPSColumns): self.columns.append(column)
    def _visit_rhs(self, rhs: MPSRhs): self.rhs.append(rhs)
    def _visit_ranges(self, range: MPSRanges): self.ranges.append(range)
    def _visit_bounds(self, bound: MPSBounds): self.bounds.append(bound)
    def _visit_quadobj(self, quadobj: MPSQuadobj): self.quadobj.append(quadobj)
    def _visit_endata(self): self.endata = True
    def _visit_free_row(self, row: MPSRow):
        # more than one free row was specified
        if self.free_row is not None:
            warn(
                f"Found a free row {RowsType.FREE_ROW} identified as '{row.identifier()}', "
                f"but one was already identified with: '{self.free_row.identifier()}'. Ignoring the new one"
            )
        else:
            self.free_row = row


# Copyright (c) 2025, InfinityQ Technology, Inc.

from abc import ABC, abstractmethod
from typing import Any, Iterable, List, Tuple, Union

import numpy as np

from titanq._model.array.arraybuilder import ArrayBuilder


class ArrayLike(ArrayBuilder, ABC):
    """
    ArrayLike represents an array like passed in by a user. We support different formats (e.g. sparse, dense),
    so this class will hide the specific format and allow different array formats to behave the same.
    """

    @abstractmethod
    def inner(self) -> Any:
        """Returns the data of the instance itself of the array like."""
        pass

    @abstractmethod
    def ndim(self) -> int:
        """Returns the dimension count of the array like"""
        pass

    @abstractmethod
    def data_type(self) -> np.dtype:
        """
        Returns the data type of the elements contained in the ArrayBuilder.

        Currently, this is represented using numpy's dtype.
        """
        pass

    @abstractmethod
    def is_binary(self) -> bool:
        """Returns if the array like has only 0s and 1s values (binary)."""
        pass

    @abstractmethod
    def sum(self, axis=None) -> Union[int, np.ndarray]:
        """Returns the sum of the array elements."""
        pass

    @abstractmethod
    def to_bytes(self) -> bytes:
        """Returns the np type array as bytes."""
        pass

    @abstractmethod
    def reshape_to_2d(self) -> None:
        """
        Reshapes the array to a 2D array.

        ex: [1, 2] ->> [[1, 2]]
        """
        pass

    @abstractmethod
    def repeat_rows(self) -> None:
        """
        Reshapes the array to a 2D array and repeats the rows.

        ex: [1, 2] --> [[1, 1], [2, 2]]
        """
        pass

    @abstractmethod
    def isnan(self) -> bool:
        """True if one of the array elements is not a number."""
        pass

    @abstractmethod
    def isinf(self) -> bool:
        """True if one of the array elements is infinity."""
        pass

    @abstractmethod
    def is_first_col_lower_second_col(self) -> bool:
        """
        Can only be used with arrays shaped as (N, 2).

        Will return if every row has their first value lower
        than the second value.

        ex: [[0, 1], [0, 1]] -> True
        ex: [[0, 0], [0, 1]] -> True
        ex: [[1, 0], [0, 1]] -> False
        """
        pass

    @abstractmethod
    def are_values_unsigned_integer(self) -> bool:
        """Returns if all contained values are unsigned integer"""
        pass

    @abstractmethod
    def row_sums_under_cardinality(self, numpy_array: np.ndarray) -> Tuple[List[int], List[int]]:
        """
        Returns all rows indices that their values sum is equal/less than the given
        cardinality in the `numpy_array`.

        :param numpy_array: array to compare to the array like sum

        :return: A tuple of the equal indices and less indices
        """
        pass

    def __iter__(self) -> Iterable['ArrayLike']:
        """Iterates over the array like and will yield a new array like on each iteration."""
        raise NotImplementedError(
            f"{self.__class__.__name__}: not implemented with '__iter__'."
        )

    def iter_nonzero_row_values(self) -> Iterable[np.ndarray]:
        """
        Iterates over a 2D array like object, and will yield non null items only
        for each row into a new numpy array
        """
        raise NotImplementedError(
            f"{self.__class__.__name__}: not implemented with 'non_null_values_row()'."
        )


# Copyright (c) 2025, InfinityQ Technology, Inc.

"""
Some helping functions for the timeouts
"""

import time


def get_dynamic_wait_interval(start_time: float) -> float:
    """
    Returns a wait interval (in seconds) based on how much time
    has elapsed since `start_time`.

    [0,   10[ --> returns 0.25 seconds
    [10,  60[ --> returns 1 seconds
    [60, inf[ --> returns 5 seconds
    """
    elapsed = time.time() - start_time

    if elapsed < 10:
        return 0.25
    elif elapsed < 60:
        return 1
    return 5



# Copyright (c) 2025, InfinityQ Technology, Inc.

"""
Some helping functions to make some text human readable
"""

def bytes_size_human_readable(size: int) -> str:
    """
    Returns a human readable string from a bytes number (size).
    Will be maxed to Terabytes.

    ex: 1000 --> '1KB' or 1 500 000 --> '1.5MB'
    """
    two_power_ten = 1024.0
    for unit in ["B", "KB", "MB", "GB"]:
        if abs(size) < two_power_ten:
            return f"{size:3.1f}{unit}"
        size /= two_power_ten
    return f"{size:3.1f}TB" # return in terabytes if above


# Copyright (c) 2024, InfinityQ Technology, Inc.

from dataclasses import dataclass
from datetime import datetime, timezone
from typing import List, Optional

from titanq._api.client import Client
from titanq import __title__ as titanq_title, __version__ as titanq_version
from titanq._util.api_key import get_and_validate_api_key

"""
Functions to query information about credits for a specific account.
"""

@dataclass
class CreditDetails():
    credits: int
    start_date: datetime
    expiration_date: datetime
    credits_used: int


@dataclass
class CreditsSummary():
    """
    The summary of credits information returned by the backend on credits request.

    Attributes
    ----------
    available_credits
        A list of `CreditDetails` objects representing credits that are still valid.
    total_available_credits
        The total amount of available credits.
    upcoming_credits
        A list of `CreditDetails` objects representing credits that have not yet started.
        This can be `None` if there are no upcoming credits.
    """
    available_credits: List[CreditDetails]
    total_available_credits: int
    upcoming_credits: Optional[List[CreditDetails]]


def get_credits_summary(api_key: str = None, base_server_url: str = "https://titanq.infinityq.io") -> CreditsSummary:
    """
    Query the amount of remaining credits and summarize available and upcoming credits.

    Parameters
    ----------
    api_key
        TitanQ API key to access the service.
        If not set, it will use the environment variable ``TITANQ_API_KEY``
    base_server_url
        TitanQ API server url, default set to ``https://titanq.infinityq.io``.

    Returns
    -------
    CreditsSummary
        An object containing the list of available credits, the total available credits,
        and the list of upcoming credits (if any).

    Raises
    ------
    MissingTitanqApiKey
        If no API key is set and is also not set as an environment variable
    """
    api_key = get_and_validate_api_key(api_key)
    client = Client(base_server_url, api_key, titanq_title, titanq_version)
    response = client.get_credits()

    upcoming_credits = []
    available_credits = []
    total_available_credits = 0

    now = datetime.now(timezone.utc)
    for credit_details in response.body:
        if credit_details.start_date > now:
            upcoming_credits.append(credit_details)
            continue
        if credit_details.expiration_date > now:
            available_credits.append(credit_details)
            total_available_credits += (credit_details.credits - credit_details.credits_used)

    return CreditsSummary(
        available_credits=available_credits,
        total_available_credits=total_available_credits,
        upcoming_credits=upcoming_credits if upcoming_credits else None
    )


# Copyright (c) 2025, InfinityQ Technology, Inc.

"""
The sparse threshold strategy allows hold the logic on how ArrayLike should be change based on a density threshold.
"""
from titanq._model.array.arraybuilder import ArrayBuilder
from titanq._model.array.arraylike import ArrayLike
from titanq._model.array.arraylike.numpy_array import NumpyArray
from titanq._model.array.arraylike.scipy_coo_array import ScipyCooArray
from titanq._model.array.arraylike.scipy_csr_array import ScipyCsrArray
from titanq._model.array.transform import ArrayTransformStrategy


class SparseThresholdStrategy(ArrayTransformStrategy):
    """
    This strategy bases on the sparse threshold, meaning depending on the calculated density
    it will return an ArrayLikeConvertOptions.
    """
    def __init__(self, threshold: float):
        self._threshold = threshold

    def check(self, array: ArrayBuilder) -> ArrayLike:
        if array.density() > self._threshold:
            return NumpyArray
        else:
            # ScipyCsrArray cannot be built with 1D arrays
            # When using SciPy 1.14+, everything should be converted to ScipyCsrArray
            if len(array.shape()) == 1:
                return ScipyCooArray
            else:
                return ScipyCsrArray


# Copyright (c) 2025, InfinityQ Technology, Inc.

from datetime import datetime
from typing import List
from pydantic import BaseModel, RootModel

class CreditEntry(BaseModel):
    credits: int
    start_date: datetime
    expiration_date: datetime
    credits_used: int


class GetCreditsResponse(RootModel):
    root: List[CreditEntry]

    def __len__(self):
        return len(self.root)

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item):
        return self.root[item]


# Copyright (c) 2025, InfinityQ Technology, Inc.

from contextlib import contextmanager
from dataclasses import dataclass
from io import BytesIO
import time
from typing import Any, Callable, Generator, Generic, List, Optional, Tuple, TypeVar
import uuid
from zipfile import ZipFile

import numpy as np
from numpy.typing import NDArray

from titanq._api.http_client import (
    BasicHttpClient,
    HTTPMethod,
    HttpClient,
    HttpClientWithAuthorization,
    HttpClientWithSession,
    HttpClientWithUserAgent,
    HttpRequest,
    HttpMetadata
)
from titanq._api.model.credits import GetCreditsResponse
from titanq._api.model.history import ComputationHistoryRequest, ComputationHistoryResponse, Status
from titanq._api.model import solve
from titanq._api.model.temp_storage import TempStorageResponse
from titanq._util.timeout import get_dynamic_wait_interval
from titanq.errors import EmptyResultError, MissingFinishedStatusError, ServerError


_API_VERSION = "v1"

_SOLVE_PATH = "solve"
_COMUTATIONS_HISTORY_PATH = "computations"
_TEMP_STORAGE_PATH = "temp_storage"
_CREDITS_PATH = "credits"

DONE_STATUS = 'done'
FINISHED_STATUS = 'finished'

T = TypeVar('T')


@dataclass
class Response(Generic[T]):
    http_metadata: HttpMetadata
    body: T


class Client:
    def __init__(
        self,
        base_server_address: str,
        api_key: str,
        name: str,
        version: str,
    ) -> None:
        self._api_key = api_key

        http_client: HttpClient = BasicHttpClient(base_server_address)
        http_client = HttpClientWithAuthorization(http_client, self._api_key)
        http_client = HttpClientWithUserAgent(http_client, name, version)

        self._http_client: HttpClient = http_client

    @contextmanager
    def session(self) -> Generator[uuid.UUID, Any, Any]:
        """
        Context manager that group all request together to make it easier to debug on the backend side

        The id of the group is yield by this context manager
        """
        old_client = self._http_client
        trace_id = uuid.uuid4()

        try:
            self._http_client = HttpClientWithSession(self._http_client, str(trace_id))
            yield trace_id
        finally:
            self._http_client = old_client

    def computation_history(self, request: ComputationHistoryRequest) -> Response[ComputationHistoryResponse]:
        """ Obtain all status/history of a computation from the request itself. """
        body, metadata = self._http_client.do(HttpRequest(
            method=HTTPMethod.GET,
            url=f"/{_API_VERSION}/{_COMUTATIONS_HISTORY_PATH}",
            query_param=request,
        ))

        return Response(
            http_metadata=metadata,
            body=ComputationHistoryResponse.model_validate_json(body),
        )

    def upload_input(self, input: solve.Request.Input) -> None:
        """ Uploads the necessary input if any."""
        input.upload_missing_local_data()

    def solve_request(self, request: solve.Request) -> Response[solve.Response]:
        """ Sends the solve request."""
        body, metadata = self._http_client.do(HttpRequest(
            method=HTTPMethod.POST,
            url=f"/{_API_VERSION}/{_SOLVE_PATH}",
            body=request,
        ))

        return Response(
            http_metadata=metadata,
            body=solve.Response(solve.Response.ApiResponse.model_validate_json(body)),
        )

    def wait_for_result(
        self,
        computation_id: uuid.UUID,
        timeout_in_sec: Optional[float] = None,
        on_update: Callable[[List[Status]], None] = None,
    ):
        """
        Waits for results to be available, will return an error if any, if not will return None

        :param computation_id: computation id to be used to fetch the statuses
        :param timeout_in_sec: if set, will stop when timeout is reached, if not will not end until the status
            indicate the computation is done.
        :param on_update: on each update, a callback function will be called with all statuses

        :raises TimeoutError: timeout has been reached
        :raises FailedComputationError: the computation failed to complete
        """
        # wait until the response is uploaded, unless a timeout is set
        start_time = time.time()
        while True:
            if timeout_in_sec is not None and time.time() < start_time + timeout_in_sec:
                break

            status_response = self.computation_history(ComputationHistoryRequest(id=computation_id))
            if status_response.http_metadata.status_code != 200:
                raise ServerError(f"Computation history request returned an unexpected http status: {status_response.http_metadata.status_code}.")

            history = status_response.body

            # History should contain only 1 element since we request to filter by
            # the computation id, unless the computation_id is wrong
            if len(history) == 0:
                raise ServerError("Computation history returned no computation.")
            if len(history) > 1:
                raise ServerError("Computation history returned more than 1 computation while only a single computation was requested.")

            # if the user has provided a callback function
            if on_update is not None:
                on_update(history[0].status)

            done, finished = _status_includes_final_state(history[0].status)

            # computation is done, end the waiting
            if done:
                if not finished:
                    raise MissingFinishedStatusError()

                return

            time.sleep(get_dynamic_wait_interval(start_time))

        raise TimeoutError(f"Timeout reached ({timeout_in_sec}s) while waiting for computation to end.")

    def result_content_length(self, request: solve.Request) -> int:
        """ Returns the content length of the result """
        return request.output.location.download_content_length(request.output.result_archive_file_name)

    def download_result(self, request: solve.Request, bytes_reader: BytesIO) -> solve.Response.Result:
        """ Download a computation's response in the given request output location """
        request.output.location.download(request.output.result_archive_file_name, bytes_reader)
        if len(bytes_reader.getvalue()) == 0:
            raise EmptyResultError("The computation ended but there is no response in the output location.")

        with ZipFile(bytes_reader) as zipFile:
            return solve.Response.Result(
                result =_extract_array_from_zip(zipFile, 'result.npy'),
                metrics=_extract_json_from_zip(zipFile, 'metrics.json', solve.Response.Result.Metrics.model_validate_json),
            )

    def download_error(self, request: solve.Request, bytes_reader: BytesIO) -> solve.Response.Result.Error:
        request.output.location.download(request.output.result_archive_file_name, bytes_reader)
        if len(bytes_reader.getvalue()) == 0:
            raise EmptyResultError("The computation ended but there is no response in the output location.")

        with ZipFile(bytes_reader) as zipFile:
            return solve.Response.Result.Error(
                error=_extract_json_from_zip(zipFile, "error.json", solve.Response.Result.Error.model_validate_json).error,
            )

    def temp_storage(self) -> Response[TempStorageResponse]:
        body, metadata = self._http_client.do(HttpRequest(
            method=HTTPMethod.GET,
            url=f"/{_API_VERSION}/{_TEMP_STORAGE_PATH}",
        ))

        return Response(
            http_metadata=metadata,
            body=TempStorageResponse.model_validate_json(body),
        )

    def get_credits(self) ->Response[GetCreditsResponse]:
        body, metadata = self._http_client.do(HttpRequest(
            method=HTTPMethod.GET,
            url=f"/{_API_VERSION}/{_CREDITS_PATH}",
        ))

        return Response(
            http_metadata=metadata,
            body=GetCreditsResponse.model_validate_json(body)
        )

def _status_includes_final_state(status: List[Status]) -> Tuple[bool, bool]:
    """
    Return if the status list of a computation includes 'done' state (the final state).

    If it also returns whether the computation has succeeded or not.
    """
    done = False
    finished = False
    for s in status:
        if s.status == DONE_STATUS:
            done = True
        elif s.status == FINISHED_STATUS:
            # no finished status indicates no result were uploaded
            finished = True

    return done, finished


def _extract_json_from_zip(zipFile: ZipFile, filename: str, parse: Callable[[bytes], T]) -> T:
    """ Returns a validated json object from a zipped json file """
    try:
        with zipFile.open(filename) as f:
            return parse(f.read())
    except KeyError:
        raise RuntimeError(f"result archive does not contain {filename}")


def _extract_array_from_zip(zipFile: ZipFile, filename: str) -> NDArray[np.float32]:
    """ returns an NDArray object from a zipped .npy file """
    try:
        with zipFile.open(filename) as f:
            return np.load(f)
    except KeyError:
        raise RuntimeError(f"result archive does not contain {filename}")


# Copyright (c) 2025, InfinityQ Technology, Inc.

from datetime import datetime
from io import BytesIO
from typing import Optional

from titanq._api.model.local_file import InputFiles, LocalFile
from titanq._api.model.location import S3Location
from titanq._api.model.solve import Manifest, Request
from titanq._storage import StorageClient


class S3Storage(StorageClient):
    """Storage client using S3 bucket from AWS"""

    def __init__(self, access_key: str, secret_key: str, bucket_name: str):
        self._access_key = access_key
        self._secret_key = secret_key
        self._bucket_name = bucket_name

        timestamp = datetime.now().isoformat()
        self._remote_folder = f"titanq_sdk/{timestamp}"

    def init(self) -> None:
        self._location = S3Location(bucket_name=self._bucket_name, access_key_id=self._access_key, secret_access_key=self._secret_key)

    def request_input_builder(
        self,
        bias: BytesIO,
        variable_bounds: BytesIO,
        weights: Optional[BytesIO] = None,
        constraint_weights: Optional[BytesIO] = None,
        constraint_bounds: Optional[BytesIO] = None,
        quad_constraint_weights: Optional[BytesIO] = None,
        quad_constraint_bounds: Optional[BytesIO] = None,
        quad_constraint_linear_weights: Optional[BytesIO] = None,
        manifest: Optional[Manifest] = None,
    ) -> Request.Input:
        files = InputFiles(
            bias_file_name=LocalFile(remote_name=self._get_full_filename("bias"), data=bias),
            variable_bounds_file_name=LocalFile(remote_name=self._get_full_filename("variable_bounds"), data=variable_bounds)
        )

        if weights is not None:
            files.weights_file_name = LocalFile(remote_name=self._get_full_filename("weights"), data=weights)
        if constraint_weights is not None:
            files.constraint_weights_file_name = LocalFile(remote_name=self._get_full_filename("constraint_weights"), data=constraint_weights)
        if constraint_bounds is not None:
            files.constraint_bounds_file_name = LocalFile(remote_name=self._get_full_filename("constraint_bounds"), data=constraint_bounds)
        if quad_constraint_weights is not None:
            files.quad_constraint_weights_file_name = LocalFile(remote_name=self._get_full_filename("quad_constraint_weights"), data=quad_constraint_weights)
        if quad_constraint_bounds is not None:
            files.quad_constraint_bounds_file_name = LocalFile(remote_name=self._get_full_filename("quad_constraint_bounds"), data=quad_constraint_bounds)
        if quad_constraint_linear_weights is not None:
            files.quad_constraint_linear_weights_file_name = LocalFile(remote_name=self._get_full_filename("quad_constraint_linear_weights"), data=quad_constraint_linear_weights)

        return Request.Input(location=self._location, files=files, manifest=manifest)

    def request_output_builder(self) -> Request.Output:
        return Request.Output(
            location=self._location,
            result_archive_file_name=self._get_full_filename("result"),
        )

    def _get_full_filename(self, filename: str) -> str:
        return f"{self._remote_folder}/{filename}"


# Copyright (c) 2024, InfinityQ Technology, Inc.
import abc
import re
from typing import List, Union
from warnings import warn

from titanq.errors import MpsMalformedFileError, MpsMissingValueError

from .model import (
    MPSBounds,
    MPSColumns,
    MPSEndata,
    MPSName,
    MPSQuadobj,
    MPSRanges,
    MPSRhs,
    MPSRow,
)
from .utils import BoundsType, ColumnsType, RowsType, SectionType, UniqueList
from ._visitor import LineSection


# Following this IBM documentation, this parser currently uses spaces as delimiters
# https://www.ibm.com/docs/en/icos/22.1.0?topic=standard-records-in-mps-format

# For integer variables both ways are supported (COLUMNS via markers and in the bounds section)
# https://www.ibm.com/docs/en/icos/22.1.1?topic=extensions-integer-variables-in-mps-files

# Integer variables markers in COLUMNS
_INTEGER_START = "'INTORG'" # 'INTORG'
_INTEGER_END = "'INTEND'" # 'INTEND'

_MPS_FILE_DEFAULT_NAME = "IMPORTED_BY_TITANQ"

# common regular expressions used in the MPS parser
# Matches "C1 R1 -2.0E+10"
SINGLE_VALUE_PATTERN = re.compile(r"^(\S+)\s+(\S+)\s+([-+]?(?:\d+\.\d*|\.\d+|\d+)(?:[eE][-+]?\d+)?)\s*$")
# Matches "C1 R1 2.0 R2 3.0" | "C1 R1 -2.0E+10 R2 3.0E+10"
DOUBLE_VALUE_PATTERN = re.compile(r"^(\S+)\s+(\S+)\s+([-+]?(?:\d+\.\d*|\.\d+|\d+)(?:[eE][-+]?\d+)?)\s+(\S+)\s+([-+]?(?:\d+\.\d*|\.\d+|\d+)(?:[eE][-+]?\d+)?)\s*$")



def _failed_to_parsed_at_section_and_line(section_name: str, line_index: int):
    raise MpsMalformedFileError(
        f"Failed to parse a line in section '{section_name}' (line: {line_index})"
    )


class SectionParser(abc.ABC):
    """class interface for any section needed to be parsed"""
    def __init__(self, start_index: int, lines: List[str]) -> None:
        super().__init__()
        self._start_index = start_index
        self._lines = lines

    def parse() -> List[LineSection]:
        """parse the given lines"""


class NameParser(SectionParser):
    """parses the NAME section of an .mps file into python objects"""

    def __init__(self, start_index, lines):
        super().__init__(start_index, lines)

        # Matches "NAME" or "NAME SOMENAME"
        self._name_pattern = re.compile(r"^NAME\s*(.*?)\s*$")

    def parse(self) -> List[MPSName]:
        if len(self._lines) > 1:
            raise MpsMalformedFileError(f"Found more than 2 lines for section '{SectionType.NAME}'")

        name_pattern_match = self._name_pattern.match(self._lines[0])
        if name_pattern_match:
            value = name_pattern_match.group(1)

            # give it a default name if nothing was set
            name = value if value else _MPS_FILE_DEFAULT_NAME
            return [MPSName(name=name)]

        _failed_to_parsed_at_section_and_line(SectionType.NAME, self._start_index)

class RowsParser(SectionParser):
    """parses the ROWS section of an .mps file into python objects"""

    def __init__(self, start_index: int, lines: List[str]) -> None:
        super().__init__(start_index, lines)
        self._rows_identifier = UniqueList(f"Found duplicates in the '{SectionType.ROWS}' section")
        self._free_row = None

        # Matches "N ANYTHING" | "E ANYTHING" | "L ANYTHING" | "G ANYTHING"
        self._rows_pattern = re.compile(r"^([NELG])\s+(\S+)\s*$")

    def _set_free_row(self, free_row_identifier):
        if self._free_row is not None:
            warn(
                f"Found a free row {RowsType.FREE_ROW} identified as '{free_row_identifier}', "
                f"but one was already identified with: '{self._free_row}'. New one will be ignored"
            )
        self._free_row = free_row_identifier


    def parse(self) -> List[MPSRow]:
        # possible format --> ['N', 'OBJ']
        rows = []
        for index, line in enumerate(self._lines):
            rows_pattern_match = self._rows_pattern.match(line)
            if rows_pattern_match:
                sense = rows_pattern_match.group(1)
                row_identifier = rows_pattern_match.group(2)

                if sense == RowsType.FREE_ROW:
                    self._set_free_row(row_identifier)

                rows.append(MPSRow(sense=sense, identifier=row_identifier))
                continue

            _failed_to_parsed_at_section_and_line(SectionType.ROWS, self._start_index + index)

        if self._free_row is None:
            raise MpsMissingValueError(f"Did not find a free row '{RowsType.FREE_ROW}' in the '{SectionType.ROWS}' section")

        return rows


class ColumnParser(SectionParser):
    """parses the COLUMNS section of an .mps file into python objects"""

    def __init__(self, start_index: int, lines: List[str]) -> None:
        super().__init__(start_index, lines)
        self._type = ColumnsType.CONTINUOUS

        self._single_value_pattern = SINGLE_VALUE_PATTERN
        self._double_value_pattern = DOUBLE_VALUE_PATTERN
        # Matches "SOMENAME 'MARKER' 'INTORG'"
        self._start_int_pattern = re.compile(r"^\S+\s+'MARKER'\s+'INTORG'\s*$")
        # Matches "SOMENAME 'MARKER' 'INTEND'"
        self._end_int_pattern = re.compile(r"^\S+\s+'MARKER'\s+'INTEND'\s*$")

    def _activate_integer_mode(self): self._type = ColumnsType.INTEGER
    def _deactivate_integer_mode(self): self._type = ColumnsType.CONTINUOUS

    def _handle_marker_line(self, value: str) -> None:
        if value == _INTEGER_START:
            self._activate_integer_mode()
        elif value == _INTEGER_END:
            self._deactivate_integer_mode()

    def parse(self) -> List[MPSColumns]:
        columns = []
        for index, line in enumerate(self._lines):
            single_pattern_match = self._single_value_pattern.match(line)
            if single_pattern_match:
                columns.append(MPSColumns(
                    identifier=single_pattern_match.group(1),
                    row_identifier=single_pattern_match.group(2),
                    coeff=float(single_pattern_match.group(3)),
                    type=self._type
                ))
                continue

            double_pattern_match = self._double_value_pattern.match(line)
            if double_pattern_match:
                columns.append(MPSColumns(
                    identifier=double_pattern_match.group(1),
                    row_identifier=double_pattern_match.group(2),
                    coeff=float(double_pattern_match.group(3)),
                    type=self._type
                ))
                columns.append(MPSColumns(
                    identifier=double_pattern_match.group(1),
                    row_identifier=double_pattern_match.group(4),
                    coeff=float(double_pattern_match.group(5)),
                    type=self._type
                ))
                continue

            start_int_pattern_match = self._start_int_pattern.match(line)
            if start_int_pattern_match:
                self._handle_marker_line(_INTEGER_START)
                continue

            end_int_pattern_match = self._end_int_pattern.match(line)
            if end_int_pattern_match:
                self._handle_marker_line(_INTEGER_END)
                continue

            _failed_to_parsed_at_section_and_line(SectionType.COLUMNS, self._start_index + index)

        return columns


class RhsParser(SectionParser):
    """parses the RHS section of an .mps file into python objects"""

    def __init__(self, start_index, lines):
        super().__init__(start_index, lines)

        self._single_value_pattern = SINGLE_VALUE_PATTERN
        self._double_value_pattern = DOUBLE_VALUE_PATTERN

    def parse(self) -> List[MPSRhs]:
        rhs = []
        for index, line in enumerate(self._lines):
            single_pattern_match = self._single_value_pattern.match(line)
            if single_pattern_match:
                rhs.append(MPSRhs(
                    identifier=single_pattern_match.group(1),
                    row_identifier=single_pattern_match.group(2),
                    coeff=float(single_pattern_match.group(3))
                ))
                continue

            double_pattern_match = self._double_value_pattern.match(line)
            if double_pattern_match:
                rhs.append(MPSRhs(
                    identifier=double_pattern_match.group(1),
                    row_identifier=double_pattern_match.group(2),
                    coeff=float(double_pattern_match.group(3))
                ))
                rhs.append(MPSRhs(
                    identifier=double_pattern_match.group(1),
                    row_identifier=double_pattern_match.group(4),
                    coeff=float(double_pattern_match.group(5))
                ))
                continue

            _failed_to_parsed_at_section_and_line(SectionType.ROWS, self._start_index + index)

        return rhs


class RangesParser(SectionParser):
    """parses the RANGES section of an .mps file into python objects"""

    def __init__(self, start_index, lines):
        super().__init__(start_index, lines)

        self._single_value_pattern = SINGLE_VALUE_PATTERN
        self._double_value_pattern = DOUBLE_VALUE_PATTERN

    def parse(self) -> List[MPSRanges]:
        ranges = []
        for index, line in enumerate(self._lines):
            single_pattern_match = self._single_value_pattern.match(line)
            if single_pattern_match:
                ranges.append(MPSRanges(
                    identifier=single_pattern_match.group(1),
                    row_identifier=single_pattern_match.group(2),
                    coeff=float(single_pattern_match.group(3))
                ))
                continue

            double_pattern_match = self._double_value_pattern.match(line)
            if double_pattern_match:
                ranges.append(MPSRanges(
                    identifier=double_pattern_match.group(1),
                    row_identifier=double_pattern_match.group(2),
                    coeff=float(double_pattern_match.group(3))
                ))
                ranges.append(MPSRanges(
                    identifier=double_pattern_match.group(1),
                    row_identifier=double_pattern_match.group(4),
                    coeff=float(double_pattern_match.group(5))
                ))
                continue

            _failed_to_parsed_at_section_and_line(SectionType.RANGES, self._start_index + index)

        return ranges


class BoundParser(SectionParser):
    """parses the BOUNDS section of an .mps file into python objects"""
    def __init__(self, start_index, lines):
        super().__init__(start_index, lines)

        # Matches "FR A B" "LO A B -2.0E+10", the first item needs to be a valid bounds option, see 'BoundsType'
        self._bounds_pattern = re.compile(r"^(LO|LI|UP|UI|FX|FR|MI|PL|BV|SC)\s+(\S+)\s+(\S+)(\s+[-+]?(?:\d+\.\d*|\.\d+|\d+)(?:[eE][-+]?\d+)?)?\s*$")

    def parse(self) -> List[MPSBounds]:
        bounds = []
        for index, line in enumerate(self._lines):
            bounds_pattern_match = self._bounds_pattern.match(line)
            if bounds_pattern_match:
                value = bounds_pattern_match.group(4)
                bounds.append(MPSBounds(
                    identifier=bounds_pattern_match.group(2),
                    type=BoundsType(bounds_pattern_match.group(1)),
                    column_identifier=bounds_pattern_match.group(3),
                    value=float(value) if value else None
                ))
                continue

            _failed_to_parsed_at_section_and_line(SectionType.BOUNDS, self._start_index + index)

        return bounds


class QuadobjParser(SectionParser):
    """parses the QUADOBJ or QMATRIX section of an .mps file into python objects"""

    def __init__(self, start_index: int, lines: List[str], type: Union[SectionType.QUADOBJ, SectionType.QMATRIX]) -> None:
        super().__init__(start_index, lines)
        self._type = type

        self._quadobj_pattern = SINGLE_VALUE_PATTERN

    def parse(self) -> List[MPSQuadobj]:
        quadobj = []
        for index, line in enumerate(self._lines):
            quadobj_pattern_match = self._quadobj_pattern.match(line)
            if quadobj_pattern_match:
                row = quadobj_pattern_match.group(1)
                column = quadobj_pattern_match.group(2)
                value = quadobj_pattern_match.group(3)

                quadobj.append(MPSQuadobj(row_identifier=row, column_identifier=column, value=float(value)))

                # if QUADOBJ is set instead of QMATRIX, write another line but inversed to fill the matrix symmetrically
                if self._type == SectionType.QUADOBJ and row != column:
                    quadobj.append(MPSQuadobj(row_identifier=column, column_identifier=row, value=float(value)))
                continue

            _failed_to_parsed_at_section_and_line(self._type, self._start_index + index)

        return quadobj


class EndataParser(SectionParser):
    """parses the ENDATA section of an .mps file into a python objects"""

    def parse(self) -> List[MPSEndata]:
        # possible format --> ['ENDATA'] # no need to parse here
        if len(self._lines) > 1:
            raise MpsMalformedFileError(f"Found more than 2 lines for section '{SectionType.ENDATA}'")

        return [MPSEndata()]


# Copyright (c) 2025, InfinityQ Technology, Inc.

from io import BytesIO
from typing import List, Optional
from titanq._api.client import Client
from titanq._api.model.local_file import InputFiles, LocalFile
from titanq._api.model.location import TempStorageLocation
from titanq._api.model.solve import Manifest, Request
from titanq._api.model.temp_storage import DualUrl
from titanq._storage import StorageClient


class TempStorage(StorageClient):
    """Temp storage client using the temp storage from InfinityQ"""

    def __init__(self, titanq_client: Client):
        self._titanq_client = titanq_client

    def init(self) -> None:
        self._tempstorage = self._titanq_client.temp_storage()

    def request_input_builder(
        self,
        bias: BytesIO,
        variable_bounds: BytesIO,
        weights: Optional[BytesIO] = None,
        constraint_weights: Optional[BytesIO] = None,
        constraint_bounds: Optional[BytesIO] = None,
        quad_constraint_weights: Optional[BytesIO] = None,
        quad_constraint_bounds: Optional[BytesIO] = None,
        quad_constraint_linear_weights: Optional[BytesIO] = None,
        manifest: Optional[Manifest] = None,
    ) -> Request.Input:
        temp_storage_input = self._tempstorage.body.input

        files = InputFiles(
            bias_file_name=LocalFile(remote_name=temp_storage_input.bias.download, data=bias),
            variable_bounds_file_name=LocalFile(remote_name=temp_storage_input.variable_bounds.download, data=variable_bounds)
        )
        urls: List[DualUrl] = [temp_storage_input.bias, temp_storage_input.variable_bounds]

        if weights is not None:
            files.weights_file_name = LocalFile(remote_name=temp_storage_input.weights.download, data=weights)
            urls.append(temp_storage_input.weights)

        if constraint_weights is not None:
            files.constraint_weights_file_name = LocalFile(remote_name=temp_storage_input.constraint_weights.download, data=constraint_weights)
            urls.append(temp_storage_input.constraint_weights)

        if constraint_bounds is not None:
            files.constraint_bounds_file_name = LocalFile(remote_name=temp_storage_input.constraint_bounds.download, data=constraint_bounds)
            urls.append(temp_storage_input.constraint_bounds)

        if quad_constraint_weights is not None:
            files.quad_constraint_weights_file_name = LocalFile(remote_name=temp_storage_input.quad_constraint_weights.download, data=quad_constraint_weights)
            urls.append(temp_storage_input.quad_constraint_weights)

        if quad_constraint_bounds is not None:
            files.quad_constraint_bounds_file_name = LocalFile(remote_name=temp_storage_input.quad_constraint_bounds.download, data=quad_constraint_bounds)
            urls.append(temp_storage_input.quad_constraint_bounds)

        if quad_constraint_linear_weights is not None:
            files.quad_constraint_linear_weights_file_name = LocalFile(remote_name=temp_storage_input.quad_constraint_linear_weights.download, data=quad_constraint_linear_weights)
            urls.append(temp_storage_input.quad_constraint_linear_weights)

        return Request.Input(location=TempStorageLocation(urls), files=files, manifest=manifest)

    def request_output_builder(self) -> Request.Output:
        return Request.Output(
            location=TempStorageLocation([self._tempstorage.body.output.result]),
            result_archive_file_name=self._tempstorage.body.output.result.upload
        )


# Copyright (c) 2025, InfinityQ Technology, Inc.

"""
Function helpers related to the Titanq api key
"""

# the official environment variable that is used in official documentation
import os
from typing import Optional

from titanq import errors


TITANQ_API_KEY_ENV = "TITANQ_API_KEY"


def get_and_validate_api_key(api_key: Optional[str]) -> str:
    """
    Validate the TitanQ's api_key. If the api_key argument is None, will try to fetch it from
    the environment variables.

    :return: The api key is validated

    :raises: MissingTitanQApiKey or InvalidTitanqApiKey
    """
    api_key = api_key or os.getenv(TITANQ_API_KEY_ENV)

    if api_key is None:
        raise errors.MissingTitanqApiKey(
            "No API key is provided. You can set your API key in the Model, "
            + "or you can set the environment variable TITANQ_API_KEY")

    return api_key


# Copyright (c) 2025, InfinityQ Technology, Inc.

import numpy as np
from scipy.sparse import coo_array, csr_array

from titanq._model.array import _TITANQ_DTYPE
from titanq._model.array.arraybuilder import ArrayBuilder


def array_builder_to_numpy(array: ArrayBuilder) -> np.ndarray:
    numpy_array = np.zeros(array.shape(), dtype=_TITANQ_DTYPE)
    if len(array.shape()) == 1:
        for (_, y), value in array.non_null_items():
            numpy_array[y] = value
    else:
        for (x, y), value in array.non_null_items():
            numpy_array[x][y] = value
    return numpy_array


def array_builder_to_coo_array(array: ArrayBuilder) -> coo_array:
    rows = []
    cols = []
    data = []
    for (x, y), value in array.non_null_items():
        rows.append(x)
        cols.append(y)
        data.append(value)

    if len(array.shape()) == 1:
        numpy_array = np.zeros(array.shape()[0])
        for value in zip(cols, data):
            numpy_array[cols] = data

        # 1D coo_array can only be created through a list or a numpy object
        return coo_array(numpy_array, shape=array.shape(), dtype=_TITANQ_DTYPE)

    return coo_array((data, (rows, cols)), shape=array.shape(), dtype=_TITANQ_DTYPE)


def array_builder_to_csr_array(array: ArrayBuilder) -> csr_array:
    # convert the ArrayBuilder to a coo_array first, since it's naturally in coordinate format.
    # coo_array is efficient for incremental construction. Converting to csr_array is very efficient.
    coo_array = array_builder_to_coo_array(array)
    return coo_array.tocsr()


# Copyright (c) 2025, InfinityQ Technology, Inc.

import copy
from typing import Any, Callable, Dict, Generic, Type, TypeVar, Union
from urllib.parse import ParseResult, urlparse

from numpy.typing import NDArray
from scipy.sparse import sparray
from pydantic import GetCoreSchemaHandler, PydanticSchemaGenerationError
from pydantic_core import core_schema
import wrapt


T = TypeVar('T')

class Obfuscated(Generic[T], wrapt.ObjectProxy):
    """
    Obfuscate is a class made to be use as base class to make the child class
    obfuscated when stringified but still serialized normally by pydantic dump functions
    """
    def __init_subclass__(cls, obfuscation: Union[str, Callable[[T], str]] = "<hidden>"):
        # Get parameter passed in [].
        # the first one is the base type that will be obfuscated on print
        base = next(base.__args__[0] for base in cls.__orig_bases__ if hasattr(base, "__args__"))

        # define both __str__ and __repr__ so the data is obfuscated
        if isinstance(obfuscation, Callable):
            def obfuscation_func(self):
                return obfuscation(self.__wrapped__)
        else:
            def obfuscation_func(_):
                return obfuscation

        setattr(cls, "__str__", obfuscation_func)
        setattr(cls, "__repr__", obfuscation_func)

        # This is related to pydantic
        # It make sure the value get cast into an Obfuscated version after it has been initialized.
        # This mean a user can pass an instance of the base class (e.g. str instead of SecretStr)
        # It also make sure the full value is dump by pydantic.
        @classmethod
        def __get_pydantic_core_schema__(cls, source: Type[Any], handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
            try:
                schema = handler(base)
            except PydanticSchemaGenerationError:
                schema = handler.generate_schema(base)
            return core_schema.no_info_after_validator_function(
                cls,
                schema,
                serialization=core_schema.plain_serializer_function_ser_schema(Obfuscated.non_obfuscated_value)
            )

        setattr(cls, "__get_pydantic_core_schema__", __get_pydantic_core_schema__)


    def non_obfuscated_value(self):
        return self.__wrapped__

    def __deepcopy__(self, memo):
        return type(self)(copy.deepcopy(self.__wrapped__, memo))


def _truncate_if_longer_than(threshold: int, suffix: str = ""):
    def truncate(value: str):
        if not value:
            return ''

        if len(value) <= threshold:
            return value
        return value[:threshold - len(suffix)] + suffix

    return truncate


def _truncate_url(url):
    parsed:ParseResult = urlparse(url)
    host = parsed.hostname

    truncated = f"{parsed.scheme}://{host}{parsed.path}"

    return truncated if str(url) == truncated else truncated + "..."


class StrTruncatedOnDisplay(Obfuscated[str], obfuscation=_truncate_if_longer_than(25, suffix="...")):...
class UrlTruncatedOnDisplay(Obfuscated[str], obfuscation=_truncate_url): ...
class ArrayLikeTruncatedOnDisplay(Obfuscated[Union[NDArray, sparray]], obfuscation=lambda a: f"<ArrayLike | shape={a.shape}>"): ...

class SecretJson(Obfuscated[Dict[str, Any]], obfuscation=lambda _: "<secret>"):...
class SecretStr(Obfuscated[str], obfuscation=lambda _: "<secret>"):...


# Copyright (c) 2025, InfinityQ Technology, Inc.

from pydantic import BaseModel, ConfigDict, Field


class DualUrl(BaseModel):
    """ Contains a download and an upload url """
    download: str
    upload: str


class _DualUrlIterable(BaseModel):
    """ Will create an iterator for a class composed with DualUrl's """
    def __iter__(self):
        return iter((getattr(self, f) for f in self.model_fields_set))


class TempStorageInput(_DualUrlIterable):
    weights: DualUrl = Field(alias="weights")
    bias: DualUrl = Field(alias="bias")
    variable_bounds: DualUrl = Field(alias="variable_bounds")
    constraint_weights: DualUrl = Field(alias="constraint_weights")
    constraint_bounds: DualUrl = Field(alias="constraint_bounds")
    quad_constraint_weights: DualUrl = Field(alias="quad_constraint_weights")
    quad_constraint_bounds: DualUrl = Field(alias="quad_constraint_bounds")
    quad_constraint_linear_weights: DualUrl = Field(alias="quad_constraint_linear_weights")

    # necessary for test cases
    model_config = ConfigDict(populate_by_name=True)


class TempStorageOutput(_DualUrlIterable):
    result: DualUrl = Field(alias="result")

    # necessary for test cases
    model_config = ConfigDict(populate_by_name=True)


class TempStorageResponse(BaseModel):
    input: TempStorageInput
    output: TempStorageOutput



# Copyright (c) 2025, InfinityQ Technology, Inc.

"""
Progress events are some type of events that would occur over a timeline with
some progress updates.
"""

import copy
from typing import Dict, Literal, Optional
from typing_extensions import override

from titanq._event import ProgressEvent, Event
from titanq._event.sink import ContentBuilder, DefinedProgress, Sink
from titanq._model.bytes_reader import BytesReaderWithCallback
from titanq._util.human_readables import bytes_size_human_readable


_DATA_TRANSFER_WAITING_EMOJI = ":hourglass_not_done:"
_DATA_TRANSFER_COMPLETED_EMOJI = ":hourglass:"


class UploadEvent(ProgressEvent):
    """Progress event when TitanQ is uploading some files."""

    def __init__(self, files: Dict[str, BytesReaderWithCallback]):
        self._files = { f: None for f in files.items()}
        self._defined_progress: DefinedProgress = None

    @override
    def start(self, sink: Sink) -> Event:
        self._defined_progress = sink.defined_progress()
        for key in self._files:
            file_name, reader = key
            cb = sink.content_builder()
            _data_transfer_update_content(cb, file_name, "upload", False)
            task_id = self._defined_progress.add_task(cb.build(), reader.total_size())

            reader.set_callback(
                lambda reader, task_id = task_id, cb = cb: self._defined_progress.update(
                    handle=task_id,
                    new_value=reader.tell(),
                    # we need to deepcopy the content builder as it should have no reference to the initial one
                    renderable=_data_transfer_append_size_content(copy.deepcopy(cb), reader.total_size(), reader.tell()).build()
            ))
            self._files[key] = task_id
            sink.display(self._defined_progress.content())

    @override
    def end(self, sink: Sink) -> Event:
        if self._defined_progress:
            for (file_name, reader), task_id in filter(lambda value: value[1] is not None, self._files.items()):
                update_content = sink.content_builder()
                _data_transfer_update_content(update_content, file_name, "upload", True)
                _data_transfer_append_size_content(update_content, reader.total_size())

                self._defined_progress.update(task_id, reader.total_size(), update_content.build())

            self._defined_progress.finalize()


class DownloadResultEvent(ProgressEvent):
    """Progress event when TitanQ is downloading results."""

    def __init__(self, file_name: str, total_size: int, bytes_reader: BytesReaderWithCallback):
        self._file_name = file_name
        self._total_size = total_size
        self._bytes_reader = bytes_reader

        self._defined_progress: DefinedProgress = None
        self._task_id = None

    @override
    def start(self, sink: Sink) -> Event:
        self._defined_progress = sink.defined_progress()

        cb = sink.content_builder()
        _data_transfer_update_content(cb, self._file_name, "download", False)
        self._task_id = self._defined_progress.add_task(cb.build(), self._total_size)

        self._bytes_reader.set_callback(
            lambda reader, task_id = self._task_id, cb = cb: self._defined_progress.update(
                handle=task_id,
                new_value=reader.tell(),
                renderable=_data_transfer_append_size_content(copy.deepcopy(cb), self._total_size, reader.tell()).build()
            )
        )
        sink.display(self._defined_progress.content())

    @override
    def end(self, sink: Sink) -> Event:
        if self._defined_progress is not None and self._task_id is not None:
            cb = sink.content_builder()
            _data_transfer_update_content(cb, self._file_name, "download", True)
            _data_transfer_append_size_content(cb, self._total_size)

            self._defined_progress.update(self._task_id, self._total_size, cb.build())
            self._defined_progress.finalize()


def _data_transfer_update_content(
    content_builder: ContentBuilder,
    file_name: str,
    data_transfer_type: Literal['download', 'upload'],
    finished: bool,
) -> ContentBuilder:
    """Formats the text for a data transfer event"""
    data_transfer_key_word_start = "Downloading" if data_transfer_type == "download" else "Uploading"
    data_transfer_key_word_end = "Downloaded" if data_transfer_type == "download" else "Uploaded"
    (content_builder
        .add_emoji(_DATA_TRANSFER_COMPLETED_EMOJI if finished else _DATA_TRANSFER_WAITING_EMOJI)
        .add_whitespace()
    )
    if finished:
        return content_builder.add_text(f"{file_name} file {data_transfer_key_word_end}")
    else:
        return content_builder.add_text(f"{data_transfer_key_word_start} {file_name} file...")


def _data_transfer_append_size_content(
    content_builder: ContentBuilder,
    total_bytes: int,
    current_bytes: Optional[int] = None,
) -> ContentBuilder:
    """
    Appends to data transfer text the current and total size.

    if current bytes is set:
        '(34.1MB/68.2MB)'
    if current_bytes is not set
        '(68.2MB)'
    """
    hr_total_bytes = bytes_size_human_readable(total_bytes)

    content_builder.add_whitespace()
    if current_bytes is not None:
        hr_current_bytes = bytes_size_human_readable(current_bytes)
        return content_builder.add_text(f"({hr_current_bytes}/{hr_total_bytes})")
    else:
        return content_builder.add_text(f"({hr_total_bytes})")


# Copyright (c) 2025, InfinityQ Technology, Inc.

from typing import Iterable, Tuple

import numpy as np
from scipy.sparse import csr_array

from titanq._model.array.arraybuilder import IndexKey, RealNumber
from titanq._model.array.arraylike.scipy_coo_array import ScipyCooArray


class ScipyCsrArray(ScipyCooArray):

    def __init__(self, array: csr_array):
        """
        Creates a ScipyCsrArray implementing ArrayLike

        :param array: The scipy Compressed Sparse Row array itself
        """
        self._array = array

    def non_null_items(self) -> Iterable[Tuple[IndexKey, RealNumber]]:
        # converting to coo_array is very quick and efficient, there is no need to re-implement this
        # as coo_array already handles this.
        scipy_coo_array = ScipyCooArray(self._array.tocoo())
        return iter(scipy_coo_array.non_null_items())

    def iter_nonzero_row_values(self) -> Iterable[np.ndarray]:
        if self.ndim() == 1:
            raise NotImplementedError(f"{self.__class__.__name__}: 1D arrays are not iterable with 'iter_nonzero_row_values()'")

        indptr = self._array.indptr
        data = self._array.data

        num_rows = self._array.shape[0]
        for i in range(num_rows):
            start = indptr[i]
            end = indptr[i + 1]
            nnz = data[start:end]
            yield nnz


# Copyright (c) 2025, InfinityQ Technology, Inc.

from dataclasses import dataclass
from typing import Any, Dict, List, Optional
import uuid

from numpy.typing import NDArray
from pydantic import BaseModel, ConfigDict, Field, SerializeAsAny, SerializerFunctionWrapHandler, model_serializer

from titanq._api.model._util import StrTruncatedOnDisplay
from titanq._api.model.local_file import InputFiles, LocalFile
from titanq._api.model.location import Location


class Parameters(BaseModel):
    model_config = ConfigDict(validate_assignment=True)

    variable_types: StrTruncatedOnDisplay
    timeout_in_secs: float
    beta: Optional[List[float]] = None
    coupling_mult: Optional[float] = None
    num_chains: Optional[int] = None
    num_buckets: Optional[int] = None
    num_engines: Optional[int] = None
    penalty_scaling: Optional[float] = None
    precision: Optional[str] = None
    constant_term: Optional[float] = None
    optimality_gap: Optional[float] = None
    presolve_ratio: Optional[float] = None

class Manifest(BaseModel):
    has_cardinality_constraint: bool
    has_set_partitioning_constraint: bool
    has_equality_constraint: bool
    has_inequality_constraint: bool


class Request(BaseModel):
    class Input(BaseModel):
        location: SerializeAsAny[Location]
        files: InputFiles
        manifest: Manifest

        @model_serializer(mode='wrap')
        def inplace_serialization(self, nxt: SerializerFunctionWrapHandler) -> Dict[str, Any]:
            """
            Change the serialization method so the top level key are not added

            e.g. {"top_key": {key_data}} -> {key_data}
            """
            dump: Dict[str, Any] = nxt(self)
            return {
                **dump.pop('location'),
                **dump.pop('files'),
                **dump
            }

        def upload_missing_local_data(self):
            # using model_dump() here (pydantic recommandation) is not possible since we
            # exclude the 'data' field. This is another approach in order to obtain all fields
            for field in self.files.__annotations__:
                value = getattr(self.files, field)
                if isinstance(value, LocalFile):
                    self.location.upload(value.remote_name, value.data)

                    # change value of LocalFile to simply a str since the data have been uploaded
                    # this make sure calling this method again will not upload the data twice
                    setattr(self.files, field, value.remote_name)


    class Output(BaseModel):
        location: SerializeAsAny[Location]
        result_archive_file_name: StrTruncatedOnDisplay

        @model_serializer(mode='wrap')
        def inplace_serialization(self, nxt: SerializerFunctionWrapHandler) -> Dict[str, Any]:
            dump: Dict[str, Any] = nxt(self)
            return {
                **dump.pop('location'),
                **dump,
            }


    input: Input
    output: Output
    parameters: Parameters


@dataclass
class Response:
    class ApiResponse(BaseModel):
        computation_id: uuid.UUID
        status: str
        message: str

    @dataclass
    class Result:
        class Metrics(BaseModel):
            computation_id: str
            solver_used: str
            computation_metrics: Dict[str, Any]
            parameters_used: Parameters
            problem_shape: Dict[str, Optional[List[int]]]


        class Error(BaseModel):
            error: str


        result:  Optional[NDArray] = None
        metrics: Optional[Metrics] = None
        error:   Optional[Error]   = None

    api_response: ApiResponse
    result: Optional[Result] = Field(None, exclude=True)



# Copyright (c) 2025, InfinityQ Technology, Inc.

from abc import ABC, abstractmethod
from contextlib import contextmanager
import logging
from typing import Optional

from titanq._event.sink import Sink
from titanq._event.sink.no_op import NoOpSink
from titanq._event.sink.pretty_text import PrettyTextSink
from titanq._event.sink.text import TextSink


log = logging.getLogger("TitanQ")


# GLOBALS
_banner_already_shown = False # wether the banner has already been shown
_optimization_counter = 1 # number of optimization already done


class Event(ABC):
    """A simple event that would occur at some point."""

    @abstractmethod
    def output(self, sink: Sink) -> None:
        """Output self to the sink."""


class ProgressEvent(ABC):
    """An event that indicates an event is tied to some kind of progress"""

    @abstractmethod
    def start(self, sink: Sink) -> None:
        pass

    @abstractmethod
    def end(self, sink: Sink) -> None:
        pass


class EventEmitter():
    """
    A class responsible for emitting events to subscribed output sinks.
    """

    def __init__(self, sink: Sink):
        self._sink = sink

    def start(self) -> None:
        """Starts the event emitter"""
        self._sink.start()

    def emit(self, event: Event) -> None:
        """Emit an event to the subscribed sinks."""
        try:
            event.output(self._sink)
        except Exception as ex:
            log.debug(f"Event failed to emit with {type(self._sink)}", ex)

    @contextmanager
    def emit_with_progress(self, event: ProgressEvent):
        try:
            event.start(self._sink)
        except Exception as ex:
            log.debug(f"Progress event failed to start with {type(self._sink)}", ex)

        yield

        try:
            event.end(self._sink)
        except Exception as ex:
            log.debug(f"Progress event failed to end with {type(self._sink)}", ex)

    def stop(self) -> None:
        """Stops the event emitter"""
        self._sink.stop()


def create_sink(log_mode: Optional[str], is_optimization: bool) -> EventEmitter:
    """Setup all sinks based on the given mode and returns the EventEmitter"""
    # having it as a lazy import fixes the circular import
    from titanq import __version__ as titanq_version

    global _banner_already_shown
    global _optimization_counter

    start_separator = f"Optimization # {_optimization_counter}" if is_optimization else None

    if log_mode in (None, "off"):
        sink = NoOpSink()
    elif log_mode == "pretty":
        sink = PrettyTextSink(version=titanq_version, show_header=not _banner_already_shown, start_separator=start_separator)
    elif log_mode == "text":
        sink = TextSink(version=titanq_version, show_header=not _banner_already_shown, start_separator=start_separator)
    else:
        raise ValueError(f"Unsupported log_mode: {log_mode}")

    # update global variables
    _banner_already_shown = True
    if is_optimization:
        _optimization_counter += 1

    return EventEmitter(sink)


# Copyright (c) 2024, InfinityQ Technology, Inc.

from typing import List
from titanq._model.constraints.constraint_type import ConstraintType


class ManifestBuilder:
    """
    A manifest builder to keep track of some data/information about the shape or the type
    that can be useful for the TitanQ API (input --> manifest).

    Currently it only holds constraint types information. Some flags will be raised in the
    manifest depending on which type of constraints are included in the request.
    """

    def __init__(self) -> None:
        self._constraints_type: List[ConstraintType] = []

    def add_constraint_type(self, constraint_type: ConstraintType) -> None:
        """ Declares the passed 'constraint_type' is being used. """
        if constraint_type is None:
            raise TypeError("Got a 'constraint_type' of type None")

        self._constraints_type.append(constraint_type)

    def has_set_partitioning_constraint(self) -> bool:
        """
        :return: if the manifest has set partitioning constraint.
        """
        return ConstraintType.SET_PARTITIONING in self._constraints_type

    def has_cardinality_constraint(self) -> bool:
        """
        :return: if the manifest has cardinality constraint.
        """
        return ConstraintType.CARDINALITY in self._constraints_type

    def has_equality_constraint(self) -> bool:
        """
        :return: if the manifest has equality constraint.
        """
        return ConstraintType.EQUALITY in self._constraints_type

    def has_inequality_constraint(self) -> bool:
        """
        :return: if the manifest has inequality constraint.
        """
        return ConstraintType.INEQUALITY in self._constraints_type



# Copyright (c) 2025, InfinityQ Technology, Inc.

"""
Array Transform module

This module provides functionality for checking various types of arrays and transform them in need.
It transforms them into different formats based on the chosen strategy.

It utilizes the strategy pattern, to provide a flexible approach for handling different array formats
and different check logic.

Example Usage
-------------
    factory = ArrayLikeFactory()

    # choose the strategy
    sprase_strategy = SparseStrategy()

    # create the transform checker with the strategy
    transform_checker = ArrayTransformChecker(sprase_strategy)

    # create the array
    array_builder = factory.create_numpy(np.array([[1, 0, 0], [0, 1, 0], [0, 0, 0]]))

    # use the transform checker
    convert_option = transform_checker.check(array)

    # convert_option is the type that you need to transform to
"""

from abc import ABC, abstractmethod

from titanq._model.array.arraybuilder import ArrayBuilder
from titanq._model.array.arraylike import ArrayLike


class ArrayTransformStrategy(ABC):
    """
    Abstract class that each strategy would implement.
    """

    @abstractmethod
    def check(self, array: ArrayBuilder) -> ArrayLike:
        """
        Checks the array if a transform is needed.

        :param array: The array to check if a transform is needed

        :returns: The type of array to transform to
        """
        pass


class ArrayTransformChecker:
    """Array transform checker class that would apply a given strategy."""

    def __init__(self, strategy: ArrayTransformStrategy):
        self._strategy = strategy


    def check(self, array: ArrayBuilder) -> ArrayLike:
        return self._strategy.check(array)


# Copyright (c) 2024, InfinityQ Technology, Inc.
from typing import Tuple
import numpy as np
from warnings import warn

from ...._model.model import Model, Vtype

from .utils import SetOnceOrDefault


class VariableBounds:
    def __init__(self, def_lower: float, def_upper: float, def_vtype: Vtype) -> None:
        """
        helper class that intialize a variable bounds with some default values and all
        of these values can only be set one
        """
        self._lower = SetOnceOrDefault(def_lower, error_msg="Variable with more than one lower bound")
        self._upper = SetOnceOrDefault(def_upper, error_msg="Variable with more than one upper bound")
        self._vtype = SetOnceOrDefault(def_vtype, allow_same_value=True, error_msg="Variable with more than one type")

    def set(self, lower: float = None, upper: float = None, vtype: Vtype = None) -> None:
        if lower is not None:
            self._lower.set(lower)

        if upper is not None:
            self._upper.set(upper)
            # if the upperbound is set to something less than 0
            if upper < 0 and self._lower.is_default():
                self._lower.set(-np.nan)
                warn(f"Found an upper bound lower than 0: '{upper}', the lower bound was set to '-np.nan'")

        if vtype is not None:
            self._vtype.set(vtype)


    def is_binary(self) -> bool:
        """ return if the variable type has been defined as a binary """
        return self._vtype.get() == Vtype.BINARY


    def into_model_variable(
        self,
        model: Model,
        name: str,
        variable_is_integer: bool,
    ) -> None:
        """vtype is optional here, it will be considered only if applied"""
        # if the variable was overriden and no bounds were set
        # we assume the bounds as 0 (zero) and 1 (one)
        if variable_is_integer and self._lower.is_default() and self._upper.is_default():
            variable_bounds = [[0.0, 1]]
        else:
            variable_bounds=[[self._lower.get(), self._upper.get()]] if self._vtype.get() is not Vtype.BINARY else None

        model.add_variable_vector(
            name=name,
            size=1,
            vtype=Vtype.INTEGER if variable_is_integer else self._vtype.get(),
            variable_bounds=variable_bounds
        )


class ConstraintBounds:
    def __init__(self, def_lower: float, def_upper: float) -> None:
        """
        helper class that intialize a constraint bounds with some default values and all
        of these values can only be set one
        """
        self._lower = SetOnceOrDefault(def_lower, error_msg="Constraint with more than one lower bound")
        self._upper = SetOnceOrDefault(def_upper, error_msg="Constraint with more than one upper bound")

    def set(self, lower: float = None, upper: float = None, vtype: Vtype = None):
        if lower is not None:
            self._lower.set(lower)

        if upper is not None:
            self._upper.set(upper)

    def get(self) -> Tuple[float, float]:
        return (self._lower.get(), self._upper.get())


# Copyright (c) 2025, InfinityQ Technology, Inc.

from typing import Iterable, List, Tuple

from titanq._model.array.arraybuilder import ArrayBuilder, IndexKey, RealNumber


class ArrayBuilderCollection(ArrayBuilder):
    """
    ArrayBuilderCollection is an implementation of the ArrayBuilder interface that contains a
    list of other ArrayBuilder instances, effectively stacking them together and ordered.
    """

    def __init__(self):
        self._collection: List[ArrayBuilder] = []

    def shape(self) -> Tuple[int, ...]:
        # We need to iterate here through the collection and manually calculate the shape
        rows_counter = 0
        last_col_shape = 0
        for item in self._collection:
            rows_counter += item.shape()[0]
            last_col_shape = item.shape()[1]

        return (rows_counter, last_col_shape)

    def density(self) -> float:
        if len(self._collection) == 0:
            return 0

        # density needs to be calculated manually. In summary, each item
        # has a weight (number of rows), we multiply the weight by it's actual
        # density. Total is the sum of density by the total number of rows.
        density = 0
        rows_counter = 0
        for item in self._collection:
            # density for each array is it's density itself mulitiplied by a weight
            # which is the number of rows
            rows = item.shape()[0]
            density += item.density() * rows
            rows_counter += rows

        return density / rows_counter

    def non_null_items(self) -> Iterable[Tuple[IndexKey, RealNumber]]:
        offset = 0
        for arr in self._collection:
            for (x, y), value in arr.non_null_items():
                yield (x + offset, y), value
            offset += arr.shape()[0]

    def is_empty(self) -> bool:
        """Returns True if no value is stored, otherwise False."""
        return len(self._collection) == 0

    def append(self, array: ArrayBuilder) -> None:
        """Appends a new array builder object to the collection."""
        if not isinstance(array, ArrayBuilder):
            raise ValueError(
                f"{self.__class__.__name__}: append() only accepts 'ArrayBuilder' objects"
            )

        if len(array.shape()) != 2:
            raise ValueError(f"{self.__class__.__name__} should only be used with 2D Arrays")

        self._collection.append(array)


# Copyright (c) 2025, InfinityQ Technology, Inc.

import io
from typing import Iterable, List, Tuple, Union

import numpy as np

from titanq._model.array.arraybuilder import IndexKey, RealNumber
from titanq._model.array.arraylike import ArrayLike, util


class NumpyArray(ArrayLike):

    def __init__(self, array: np.ndarray):
        """
        Creates a NumpyArray implementing ArrayLike

        :param array: The numpy ndarray itself
        """
        self._array = array

    def shape(self) -> Tuple[int, int]:
        return self._array.shape

    def density(self) -> float:
        return np.count_nonzero(self._array) / self._array.size

    def non_null_items(self) -> Iterable[Tuple[IndexKey, RealNumber]]:
        non_zero_indices = np.nonzero(self._array)
        if len(non_zero_indices) == 1: # 1D
            for i in non_zero_indices:
                yield (0, i), self._array[i]
        else: # 2D
            for i, j in zip(non_zero_indices[0], non_zero_indices[1]):
                yield (i, j), self._array[i, j]

    def inner(self) -> np.ndarray:
        return self._array

    def ndim(self) -> int:
        return self._array.ndim

    def data_type(self) -> np.dtype:
        return self._array.dtype

    def is_binary(self) -> bool:
        return np.logical_xor((self._array == 1), (self._array == 0)).all()

    def sum(self, axis=None) -> Union[int, np.ndarray]:
        return np.sum(self._array, axis=axis)

    def to_bytes(self) -> bytes:
        buffer = io.BytesIO()
        np.save(buffer, self._array)
        return buffer.getvalue()

    def reshape_to_2d(self) -> None:
        if self.ndim() != 1:
            raise ValueError("Input array is not 1-D. Cannot reshape array.")

        self._array = self._array.reshape(1, -1)

    def repeat_rows(self) -> None:
        self._array = np.repeat(self._array, 2).reshape(-1, 2)

    def isnan(self) -> bool:
        return np.isnan(self._array).any()

    def isinf(self) -> bool:
        return np.isinf(self._array).any()

    def __iter__(self) -> Iterable[Union[RealNumber, 'NumpyArray']]:
        if self.ndim() == 1: # check for ndim only once for performance gains
            for value in self._array:
                yield value
        else:
            for row in self._array:
                yield NumpyArray(row)

    def iter_nonzero_row_values(self) -> Iterable[np.ndarray]:
        if self.ndim() == 1:
            raise NotImplementedError(f"{self.__class__.__name__}: 1D arrays are not iterable with 'iter_nonzero_row_values()'")

        for row in self._array:
            yield row[row != 0]

    def is_first_col_lower_second_col(self) -> bool:
        return util.compare_first_col_lower_second_col(self._array)

    def are_values_unsigned_integer(self) -> bool:
        return util.are_values_unsigned_integer(self._array)

    def row_sums_under_cardinality(self, numpy_array: np.ndarray) -> Tuple[List[int], List[int]]:
        return util.row_sums_under_cardinality(self.sum(axis=1), numpy_array)


# Copyright (c) 2024, InfinityQ Technology, Inc.
from .model import Model
from .objective import Target
from .optimize_response import OptimizeResponse
from .variables import Vtype
from .precision import Precision

# constraints folder
from .constraints import Constraints, QuadConstraints


# keep fast_sum and fastSum for retro compatibility. mark them as deprecated
from deprecated import deprecated

@deprecated(reason="use Python's built-in sum function instead")
def fast_sum(*args, **kwargs):
    return sum(*args, **kwargs)

@deprecated(reason="use Python's built-in sum function instead")
def fastSum(*args, **kwargs):
    return sum(*args, **kwargs)


# Copyright (c) 2025, InfinityQ Technology, Inc.

from abc import ABC, abstractmethod
from typing import Iterable, Tuple, Union


IndexKey = Tuple[int, int]
RealNumber = Union[int, float]


class ArrayBuilder(ABC):
    """
    ArrayBuilder allows data structures to be eventually converted into an ArrayLike.
    """

    @abstractmethod
    def density(self) -> float:
        """Returns the density of the array builder."""
        pass

    @abstractmethod
    def shape(self) -> Tuple[int, ...]:
        """Returns the shape of the array builder."""
        pass

    @abstractmethod
    def non_null_items(self) -> Iterable[Tuple[IndexKey, RealNumber]]:
        """
        Iterate over the non-null content of the array builder.

        :param start_row_index: appends it to the start of the row index, this is useful when iterating
        over multiple array builders.
        """
        pass


# Copyright (c) 2024, InfinityQ Technology, Inc.
import abc


class Visitor(abc.ABC):
    """Visitor class containing all visit methods"""
    def _visit_name(self, name):
        pass

    def _visit_rows(self, row):
        pass

    def _visit_free_row(self, row):
        pass

    def _visit_columns(self, column):
        pass

    def _visit_rhs(self, rhs):
        pass

    def _visit_ranges(self, range):
        pass

    def _visit_bounds(self, bound):
        pass

    def _visit_quadobj(self, quadobj):
        pass

    def _visit_endata(self):
        pass

class LineSection(abc.ABC):
    """This is the base class for a line of a section of the MPS file"""

    def _accept(self, visitor: Visitor):
        pass


# Copyright (c) 2025, InfinityQ Technology, Inc.

from dataclasses import dataclass
from typing import Dict


@dataclass
class AwsCredentials:
    """
    AWSCredentials is a data class that stores the credentials required
    to authenticate and interact with Amazon Web Service (AWS).
    """
    access_key_id: str
    secret_access_key: str


@dataclass
class GcpCredentials:
    """
    GCPCredentials is a data class that stores the credentials required
    to authenticate and interact with Google Cloud Platform (GCP) services.
    """
    json_key: Dict[str, str]


# Copyright (c) 2025, InfinityQ Technology, Inc.

"""
Progress events are some type of events that would occur over a timeline with
some progress updates.
"""

from typing import Any, List, Optional, TypeVar
from typing_extensions import override

from titanq._api.model.history import Status
from titanq._event import Event, ProgressEvent
from titanq._event.sink import ContentBuilder, Sink, UndefinedProgress

T = TypeVar('T')


_WAITING_UNDEFINED_PROGRESS_COLOR = "orange3"


class PreparingDataEvent(ProgressEvent):
    """Undefined progress event when the SDK is preparing data."""

    def __init__(self):
        self._undefined_progress: UndefinedProgress = None

    @override
    def start(self, sink: Sink) -> Event:
        self._undefined_progress = sink.undefined_progress()
        sink.display(
            self._undefined_progress.update(
                renderable=self._preparing_content_builder(sink)
                    .set_color(_WAITING_UNDEFINED_PROGRESS_COLOR)
                    .add_tailing_spinner()
                    .build()
            )
        )

    @override
    def end(self, sink: Sink) -> Event:
        sink.display(
            self._undefined_progress.update(
                renderable=self._preparing_content_builder(sink).build()
            )
        )
        sink.display(
            sink.content_builder()
                .add_emoji(":package:")
                .add_whitespace()
                .add_text("Data prepared for TitanQ")
                .build()
        )

    def _preparing_content_builder(self, sink: Sink) -> ContentBuilder:
        return (
            sink.content_builder()
                .add_emoji(":hammer:")
                .add_whitespace()
                .add_text("Preparing data for TitanQ")
        )


class SendingProblemEvent(ProgressEvent):
    """Undefined progress event when the SDK is sending a request to TitanQ."""

    def __init__(self):
        self._undefined_progress: UndefinedProgress = None

    @override
    def start(self, sink: Sink) -> Event:
        self._undefined_progress = sink.undefined_progress()
        sink.display(
            self._undefined_progress.update(
                renderable=self._preparing_content_builder(sink)
                    .set_color(_WAITING_UNDEFINED_PROGRESS_COLOR)
                    .add_tailing_spinner()
                    .build()
            )
        )

    @override
    def end(self, sink: Sink) -> Event:
        sink.display(
            self._undefined_progress.update(
                renderable=self._preparing_content_builder(sink).build()
            )
        )
        sink.display(
            sink.content_builder()
                .add_emoji(":envelope_with_arrow:")
                .add_whitespace()
                .add_text("Request received by TitanQ")
                .build()
        )

    def _preparing_content_builder(self, sink: Sink) -> ContentBuilder:
        return (
            sink.content_builder()
                .add_emoji(":rocket:")
                .add_whitespace()
                .add_text("Sending request to TitanQ")
        )


class WaitingForResultEvent(ProgressEvent):
    """Undefined progress event when the SDK is waiting after the results."""

    def __init__(self):
        self._undefined_progress: UndefinedProgress = None
        self._latest_status_list = None # keep in memory the latest
        self._sink = None

    def update_status_list(self, status_list: List[Status]):
        """Updates the status list displayed in this event"""
        if self._sink:
            self._on_status_updates(self._sink, status_list)

    @override
    def start(self, sink: Sink) -> Event:
        self._sink = sink
        self._undefined_progress = sink.undefined_progress()

    @override
    def end(self, sink: Sink) -> Event:
        sink.display(
            self._undefined_progress.update(
                renderable=sink.content_builder()
                    .add_emoji(":brain:")
                    .add_whitespace()
                    .add_text("TitanQ is working its magic")
                    .build(),
                additional_renderables=self._build_status_content(sink, self._latest_status_list)
            )
        )
        sink.display(
            sink.content_builder()
                .add_emoji(":bulb:")
                .add_whitespace()
                .add_text("Solution received from TitanQ")
                .build()
        )

    def _on_status_updates(self, sink: Sink, status_list: Optional[List[Status]] = None) -> None:
        self._latest_status_list = status_list
        sink.display(
            self._undefined_progress.update(
                renderable=sink.content_builder()
                    .set_color(_WAITING_UNDEFINED_PROGRESS_COLOR)
                    .add_emoji(":brain:")
                    .add_whitespace()
                    .add_text("TitanQ is working its magic")
                    .add_tailing_spinner()
                    .build(),
                additional_renderables=self._build_status_content(sink, self._latest_status_list, with_colors=True)
        ))

    def _build_status_content(
        self,
        sink: Sink,
        status_list: Optional[List[Status]] = None,
        with_colors: bool = False,
    ) -> List[Any]:
        if status_list is None:
            return None

        content_list = []
        for i, status in enumerate(status_list):
            content_builder = (
                sink.content_builder()
                    .add_text("Status update: ")
                    .add_text(status.status)
                    .add_whitespace()
                    .add_text("(")
                    .add_text(status.timestamp.strftime("%H:%M:%S"))
                    .add_text(")")
            )

            if with_colors:
                if "failed" in status.status:
                    content_builder.set_color("red")
                if "finished" in status.status:
                    content_builder.set_color("green")
                elif i == len(status_list) - 1: # last item
                    content_builder.set_color("blue")

            content_list.append(content_builder.build())
        return content_list


class BuildingFromMPSEvent(ProgressEvent):
    """Undefined progress event when the SDK building the model from an mps file."""

    def __init__(self, file_name: str):
        self._file_name = file_name
        self._undefined_progress: UndefinedProgress = None

    @override
    def start(self, sink: Sink) -> Event:
        self._undefined_progress = sink.undefined_progress()
        sink.display(
            self._undefined_progress.update(
                renderable=self._preparing_content_builder(sink)
                    .set_color(_WAITING_UNDEFINED_PROGRESS_COLOR)
                    .add_tailing_spinner()
                    .build()
            )
        )

    @override
    def end(self, sink: Sink) -> Event:
        sink.display(
            self._undefined_progress.update(
                renderable=self._preparing_content_builder(sink).build()
            )
        )
        sink.display(
            sink.content_builder()
                .add_emoji(":wrench:")
                .add_whitespace()
                .add_text(f"Model built from MPS file ({self._file_name})")
                .build()
        )

    def _preparing_content_builder(self, sink: Sink) -> ContentBuilder:
        return (
            sink.content_builder()
                .add_emoji(":wrench:")
                .add_whitespace()
                .add_text(f"Loading and building model from MPS file ({self._file_name})")
        )




# Copyright (c) 2024, InfinityQ Technology, Inc.
from typing import List, Optional

from titanq.errors import MpsMissingSectionError, MpsUnexpectedValueError
from .model import MPSBounds, MPSColumns, MPSObject, MPSRanges, MPSRhs, MPSRow
from .utils import SectionType


def _get_row_identifiers(rows: List[MPSRow], free_row: Optional[MPSRow] = False) -> List[str]:
    """returns a unique list of all ROWS identifiers, will include the free row if set to True"""
    identifiers = set([row.identifier() for row in rows])
    if free_row:
        identifiers.add(free_row.identifier())
    return identifiers


def _get_columns_identifiers(columns: List[MPSColumns]) -> List[str]:
    """returns a unique list of all COLUMNS identifiers, will include the free row if set to True"""
    return set([column.identifier() for column in columns])


def validate_columns_identifiers(rows: List[MPSRow], free_row: MPSRow, columns: List[MPSColumns]):
    """
    check if any row identifier in the columns does not match an existing row
    """
    rows_identifiers = _get_row_identifiers(rows, free_row)

    for column in columns:
        if column.row_identifier() not in rows_identifiers:
            raise MpsUnexpectedValueError(
                f"Found an unknown row identifier in the section '{SectionType.COLUMNS}': '{column.row_identifier()}'"
            )


def validate_rhs_identifiers(rows: List[MPSRow], rhs: List[MPSRhs]):
    """check if any rhs identifier in the columns does not match an existing row"""
    rows_identifiers = _get_row_identifiers(rows)
    for r in rhs:
        if r.row_identifier() not in rows_identifiers:
            raise MpsUnexpectedValueError(
                f"Found an unknown row identifier in the section '{SectionType.RHS}': '{r.row_identifier()}'"
            )


def validate_ranges_identifiers(rows: List[MPSRow], ranges: List[MPSRanges]):
    """check if any row identifier in the columns does not match an existing row"""
    rows_identifiers = _get_row_identifiers(rows)
    for range in ranges:
        if range.row_identifier() not in rows_identifiers:
            raise MpsUnexpectedValueError(
                f"Found an unknown row identifier in the section '{SectionType.RANGES}': '{range.row_identifier()}'"
            )


def validate_bounds_identifiers(columns: List[MPSColumns], bounds: List[MPSBounds]):
    """check if any column identifier is the bounds does not match an existing column"""
    columns_identifiers = _get_columns_identifiers(columns)

    for bound in bounds:
        if bound.column_identifier() not in columns_identifiers:
            raise MpsUnexpectedValueError(
                f"Found an unknown column identifier in the section '{SectionType.BOUNDS}': '{bound.column_identifier()}'"
            )

def validate_missing_required_sections(mps_object: MPSObject) -> None:
    """check if any of the required section is missing"""
    missing_section = None

    # mandatory sections
    if mps_object.name is None:
        missing_section = SectionType.NAME
    elif len(mps_object.rows) == 0 :
        missing_section = SectionType.ROWS
    elif len(mps_object.columns) == 0:
        missing_section = SectionType.COLUMNS
    elif not mps_object.endata:
        missing_section = SectionType.ENDATA

    if missing_section:
        raise MpsMissingSectionError(f"Section '{missing_section}' was not found, this section is required in an .mps file")


# Copyright (c) 2025, InfinityQ Technology, Inc.

from datetime import datetime
from http import HTTPStatus
from pydantic import BaseModel
import uuid

from titanq import errors


class Problem(BaseModel):
    """
    Problem represents an RFC 7807/RFC 9457-compliant error response.
    It defines the standard structure for problem details in HTTP APIs form TitanQ.
    """
    status_code: int
    type: str
    title: str
    detail: str
    instance: str
    timestamp: datetime
    trace_id: uuid.UUID
    status_text:str

    def raise_(self) -> None:
        """Raise the appropriate error associated with this problem"""

        if self.status_code == HTTPStatus.BAD_REQUEST:
            raise errors.BadRequest(f"{self.detail}")
        elif self.status_code == HTTPStatus.PAYMENT_REQUIRED:
            raise errors.NotEnoughCreditsError()

        elif 400 <= self.status_code < 500:
            raise errors.ClientError(f"{self.detail}")

        elif self.status_code == HTTPStatus.NOT_IMPLEMENTED:
            raise errors.UnsolvableRequestError()

        elif 500 <= self.status_code < 600:
            raise errors.ServerError(f"{self.detail}")

        raise errors.UnexpectedServerResponseError(f'{self.status_code} {self.detail}')


# Copyright (c) 2025, InfinityQ Technology, Inc.

from abc import ABC, abstractmethod
from typing import Generic, List, Optional, TypeVar
from typing_extensions import Self

T = TypeVar("T")


class ContentBuilder(ABC, Generic[T]):
    """
    This class builds some content to eventually be displayed.

    Usage example:
    >>>    content_builder = (
                ConcreteContentBuilder()
                .content_builder.add_emoji("👋")
                .content_builder.add_whitespace()
                .content_builder.add_text("Hello, World!")
        )

        content_builder.build()
        # "👋 Hello, World!"
    """

    @abstractmethod
    def build(self) -> T:
        """Builds the content."""

    @abstractmethod
    def set_color(self, color: str) -> Self:
        """Sets the color of the content."""

    @abstractmethod
    def add_emoji(self, emoji: str) -> Self:
        """Add an emoji to the content."""

    @abstractmethod
    def add_text(self, text: str) -> Self:
        """Add a text to the content."""

    @abstractmethod
    def add_whitespace(self) -> Self:
        """Add a whitespace to the content."""

    @abstractmethod
    def add_tailing_spinner(self) -> Self:
        """Add a spinner animation in the end of the content."""


class DefinedProgress(ABC, Generic[T]):
    """
    This class manages a process of a defined progress.

    Usage example:
        tasks = { "some task": 500 }
        defined_progress = ConcreteDefinedProgress(tasks)
        defined_progress.update("some task", 50)

        defined_progress.finalize()
    """
    @abstractmethod
    def add_task(self, renderable: T, total: int) -> int:
        """Adds a task and return a handle to use with further methods."""

    @abstractmethod
    def content(self) -> T:
        """Returns the content of the defined progress"""

    @abstractmethod
    def update(self, handle: int, new_value: int, renderable: Optional[T] = None) -> None:
        """Update the progress builder process."""

    @abstractmethod
    def finalize(self) -> None:
        """Finalize the progress building process."""


class UndefinedProgress(ABC, Generic[T]):
    """
    This class manages a process of a defined progress.

    Usage example:
        undefined_progress = ConcreteUndefinedProgress()

        undefined_progress.update("Hello, World")
    """

    @abstractmethod
    def update(self, renderable: T, additional_renderables: Optional[List[T]] = None) -> T:
        """
        Update the undefined progress with some renderable(s).

        Optionally, the sink can take additional renderables as a list.

        :param renderable: the main renderable to be displayed
        :param additional_renderables: A list of additional renderables to be displayed
        """


class Sink(ABC, Generic[T]):
    """
    Abstract class of a Sink type object. A Sink is any output type that you could
    use to transfer information from somewhere to another destination.

    As an example, a logger is a Sink.
    """

    @abstractmethod
    def start(self) -> None:
        """Starts the sink."""

    @abstractmethod
    def stop(self) -> None:
        """Stops the sink."""

    @abstractmethod
    def display(self, renderable: T, with_timestamp: Optional[bool] = True) -> None:
        """
        Display the content from a renderable.

        :param with_timestamp: If False, will not show any timestamp
        """

    @abstractmethod
    def content_builder(self) -> ContentBuilder[T]:
        """Returns a content builder to be able to build an output."""

    @abstractmethod
    def defined_progress(self) -> DefinedProgress[T]:
        """
        Returns a defined progress to be able to build and update an output.
        Multiple tasks can be assigned to a defined progress.
        """

    @abstractmethod
    def undefined_progress(self) -> UndefinedProgress[T]:
        """
        Returns an undefined progress to be able to build an output.
        Only a single task can be assigned to an undefined progress.
        """


# Copyright (c) 2025, InfinityQ Technology, Inc.

from typing import Iterable, List, Tuple, Union

from titanq._model.array.arraybuilder import ArrayBuilder, IndexKey, RealNumber


class IndexValueArrayBuilder(ArrayBuilder):
    """
    IndexValueArrayBuilder is an implementation of the ArrayBuilder interface that stores data in a dictionary
    where the keys are tuples of indices (x, y) and the values are real numbers (int or float).

    The additional methods keys() and values() make it faster to retrieve arrays from this builder in some cases.
    """

    def __init__(self, shape: Tuple[int, ...]):
        """
        Initialize an IndexValueArrayBuilder.

        :param shape: Determine the shape of the IndexValueArrayBuilder, this is needed to calculate the density.
        """
        self._shape = shape
        self._data = {}

    def shape(self) -> Tuple[int, ...]:
        return self._shape

    def density(self) -> float:
        return len(self._data) / (self.shape()[0] * self.shape()[1])

    def non_null_items(self) -> Iterable[Tuple[IndexKey, RealNumber]]:
        return self._data.items()

    def is_empty(self) -> bool:
        """Returns True if no value is stored, otherwise False."""
        return not self._data

    def append(self, x: int, y: int, value: Union[int, float]) -> None:
        """Adds or update a value at the given (x, y) index."""
        self._data[(x, y)] = self._data.get((x, y), 0) + value

    def keys(self) -> List[IndexKey]:
        """Returns the coordinates of the stored data."""
        return self._data.keys()

    def values(self) -> List[RealNumber]:
        """Returns the values of the stored data"""
        return self._data.values()